from typing import TypeVar, Optional, Generic
from threading import Thread, Condition
from collections import deque
from queue import SimpleQueue


T = TypeVar("T")

def none_throws(optional: Optional[T], message: str = "Unexpected `None`") -> T:
    if optional is None:
        raise AssertionError(message)
    return optional


def _thread_pool_worker(id:int, exc_queue:SimpleQueue, target, args, kwargs):
    try:
        target(*args, **kwargs)
        exc_queue.put((id, None))
    except Exception as exc:
        exc_queue.put((id, exc))


class ThreadPool:
    """Threadpool that can join() all started threads at the same time, but if
    any of those threads got caught in an exception, join() will reraise that
    exception as soon as it is received. It is then up to you to stop any other
    threads. This pool will wait for them when it exits the context.
    """
    def __init__(self):
        self.threads = {}
        self.queue = SimpleQueue()

    def start(self, func, *args, **kwargs):
        thread_id = len(self.threads)
        self.threads[thread_id] = Thread(
            target=_thread_pool_worker,
            kwargs={
                "id": thread_id,
                "exc_queue": self.queue,
                "target": func,
                "args": args,
                "kwargs": kwargs,
            },
            name=func.__name__)
        self.threads[thread_id].start()

    def join(self):
        while len(self.threads) > 0:
            thread_id, exc = self.queue.get()
            self.threads[thread_id].join()
            del self.threads[thread_id]
            if exc is not None:
                raise exc

    def __enter__(self):
        return self
    
    def __exit__(self, *args, **kwargs):
        for thread in self.threads.values():
            thread.join()
        self.threads = {}


class Cancelled(Exception):
    """Error raised by CancelableQueue's `put()` or `get()` when `cancel()` was
    called.
    """
    pass


T = TypeVar('T')

class CancelableQueue(Generic[T]):
    """SimpleQueue, but when cancel() is called it will release all blocking
    put() and get() calls and raise `Cancelled`. Also much worse performance
    than SimpleQueue so don't use for heavy workloads plz.
    """
    def __init__(self, capacity:Optional[int]=None):
        self.capacity = capacity
        self.size = 0
        self.queue = deque()
        self.cv = Condition()
        self.cancelled = False

    def put(self, item: T):
        """put() blocks until there's space on the queue. Can raise `Cancelled`
        when `cancel()` was called.
        """
        with self.cv:
            self.cv.wait_for(lambda: self.cancelled or self.capacity is None or self.size < self.capacity)
            if self.cancelled:
                raise Cancelled()
            self.queue.append(item)
            self.size += 1
            self.cv.notify()

    def get(self) -> T:
        """blocking get(). Either returns an item from the queue, or raises
        `Cancelled`.
        """
        with self.cv:
            self.cv.wait_for(lambda: self.cancelled or self.size > 0)
            if self.cancelled:
                raise Cancelled()
            self.size -= 1
            item = self.queue.popleft()
            self.cv.notify()
        return item
    
    def cancel(self):
        """Makes all calls to `get()` and `put()` raise `Cancelled()`."""
        with self.cv:
            self.cancelled = True
            self.cv.notify_all()    



#!/usr/bin/env python3
'''Segments Japanese text using the fugashi tokenizer.
Note the specifics of Japanese tokenization, where verbs are always separate to stem
and  conjugation part, as well topic or subject particles are split from the nouns.
This means that the Japanese sentences would likely be quite a bit longer than the
English ones.'''
import fugashi
from sys import stdin

tagger = fugashi.Tagger()

# https://www.dampfkraft.com/nlp/how-to-tokenize-japanese.html
for line in stdin:
    line = line.strip()
    words = [word.surface for word in tagger(line)]
    print(*words)



import sys
import os
import json
from collections import deque
from opusfilter import FilterABC, PreprocessorABC, ConfigurationError
from queue import SimpleQueue
from shlex import quote
from subprocess import Popen, PIPE
from threading import Thread
from typing import Optional, Dict, Any
from xxhash import xxh32


def encode_env(type_name: str, value: Any) -> str:
    if type_name == 'bool':
        return '1' if value else ''
    else:
        return str(value)


def load_filter_definition(filter_name:str) -> Dict:
    with open('filters/{filter_name}.json') as fh:
            return json.load(fh)


def generate_filter_command(filter_definition:Dict, parameters:Dict) -> str:
    filter_definition = filters[step['filter']]

    # List of k=v shell variable definitions
    filter_params = [
            '{}={}'.format(name, quote(encode_env(props['type'], parameters.get(name, props.get('default', None)))))
            for name, props in filter_definition['parameters'].items()
    ]

    # Command, prefixed by variable definitions so they get expanded
    # correctly in the command bit.
    return '; '.join(filter_params + [filter_definition['command']])


def patch_environ() -> Optional[Dict[str,str]]:
    # Make sure the path to the python binary (and the installed utils)
    # is in the PATH variable. If you load a virtualenv this happens by
    # default, but if you call it with the virtualenv's python binary 
    # directly it wont.
    pyenv_bin_path = os.path.dirname(sys.executable)
    os_env_bin_paths = os.environ.get('PATH', '').split(os.pathsep)
    return {
            **os.environ,
            'PATH': os.pathsep.join([pyenv_bin_path] + os_env_bin_paths)
    } if pyenv_bin_path not in os_env_bin_paths else None


def feed_child_worker(input_queue:SimpleQueue, stdin):
    while True:
        line = input_queue.pop()
        if line is None:
            break
        stdin.write(line.encode() + b'\n')
    stdin.close()


def read_child_worker(stdout, output_queue:SimpleQueue):
    for line in stdout:
        output_queue.put(line.rstrip(b'\r\n').decode())
    output_queue.put(None)


class OpusCleanerPreprocessor(PreprocessorABC):
    def __init__(self, filter:str, parameters:Dict[str,Any], column:int, **kwargs):
        filter_definition = load_filter_definition(filter)
        if filter_definition['type'] != 'monolingual':
            raise ConfigurationError()

        self.command = generate_filter_command(filter_definition, parameters)
        self.column = column
        super().__init__(**kwargs)

    def process(self, pairs):
        # Single column python -> child
        input_queue = SimpleQueue()
        # Single column child -> python
        output_queue = SimpleQueue()
        # Remainder of the columns python -> python
        column_queue = deque()

        child = Popen(self.command, cwd=basedir, stdin=PIPE, stdout=PIPE, env=patch_environ())

        feeder = Thread(target=feed_child_worker, args=[input_queue, child.stdin])
        feeder.start()

        reader = Thread(target=read_child_worker, args=[child.stdout, output_queue])
        reader.start()

        def split(pair):
            column_queue.append(pair[:self.column] + pair[self.column+1:])
            return pair[self.column]

        def merge(val):
            rest = column_queue.popleft()
            return rest[:self.column] + [val] + rest[self.column:]

        for pair in pairs:
            # Push input pair
            input_queue.put(split(pair))

            # Yield any available output pairs
            while True:
                try:
                    yield merge(output_queue.get_nowait())
                except Empty:
                    break
        
        # Signal worker to stop
        input_queue.put(None)
        feeder.join()

        # Yield remaining output pairs
        while True:
            pair = output_queue.get()
            if pair is not None:
                yield merge(pair)
            else:
                break

        retval = child.wait()
        reader.join()

        if retval != 0:
            raise Exception(f'Child process {command} exited with non-zero exit code: {retval}')

        assert len(column_queue) == 0


class OpusCleanerFilter(FilterABC):
    """One Big Hack (Tm)"""

    def __init__(self, filter:str, parameters:Dict[str,Any], **kwargs):
        filter_definition = load_filter_definition(filter)
        if filter_definition['type'] != 'bilingual':
            raise ConfigurationError()

        self.command = generate_filter_command(filter_definition, parameters)
        super().__init__(**kwargs)

    def accept(self, score):
        return bool(score)

    def score(self, pairs):
        input_queue = SimpleQueue()
        output_queue = SimpleQueue()
        
        input_log = deque()
        
        child = Popen(self.command, cwd=basedir, stdin=PIPE, stdout=PIPE, env=patch_environ())

        feeder = Thread(target=feed_child_worker, args=[input_queue, child.stdin])
        feeder.start()

        reader = Thread(target=read_child_worker, args=[child.stdout, output_queue])
        reader.start()

        def record(pair):
            """Record the hash of the line so we know whether it makes it through the filter"""
            line = '\t'.join(pair)
            input_log.append(xxh32(line).digest())
            return line

        def catch_up(line):
            """Yield 0 for all the lines that got skipped by the filter"""
            key = xxh32(line).digest()
            while column_queue.popleft() != key:
                yield 0
            yield 1

        for pair in pairs:
            # Push input pair
            input_queue.put(record(pair))

            # Yield any available output pairs
            while True:
                try:
                    yield from catch_up(output_queue.get_nowait())
                except Empty:
                    break
        
        # Signal worker to stop
        input_queue.put(None)
        feeder.join()

        # Yield remaining output pairs
        while True:
            pair = output_queue.get()
            if pair is not None:
                yield from catch_up(pair)
            else:
                break

        retval = child.wait()
        reader.join()

        if retval != 0:
            raise Exception(f'Child process {command} exited with non-zero exit code: {retval}')

        assert len(column_queue) == 0



#!/usr/bin/env python3
import sys

def main():
    for line in sys.stdin:
        fields = line.strip("\r\n").split("\t")
        ok = True
        for field in fields:
            if not field.strip():
                ok = False
                break
        if ok:
            sys.stdout.write(line)

if __name__ == '__main__':
    main()



import io
import unittest

from num_mismatch import filter_numerical_mismatch

class TestNumMismatch(unittest.TestCase):
	def _test(self, line:str, ratio:float, **kwargs) -> bool:
		fin = io.StringIO(line)
		fout = io.StringIO()
		filter_numerical_mismatch(fin, fout, ratio, **kwargs)
		return fout.getvalue() == line

	def assertAccept(self, line:str, ratio:float, **kwargs):
		"""Test that this line is accepted"""
		self.assertTrue(self._test(line, ratio, **kwargs))

	def assertReject(self, line:str, ratio:float, **kwargs):
		"""Test that this line is rejected"""
		self.assertFalse(self._test(line, ratio, **kwargs))

	def test_match(self):
		"""Exact matches should be accepted."""
		self.assertAccept('There are 74 cows\t74 cows have we', 1.0)

	def test_accepted_comma_mismatch(self):
		"""Differences in the decimal separator should be accepted."""
		self.assertAccept('There are 7.4 cows\t7,4 cows have we', 1.0)

	def test_mismatch(self):
		"""Differences in the number should be rejected."""
		self.assertReject('There are 73 cows\t74 cows have we', 1.0)

	def test_ratio(self):
		"""Lowering the ratio threshold will accept mismatches"""
		line = 'There are 73 cows in 6 fields\tWe have 6 fields with 74 cows' # 2 / 3 = 0.667
		self.assertAccept(line, 0.5)
		self.assertReject(line, 1.0)

	def test_prefix_zero(self):
		"""Numbers like 06 and 6 should be the same. See #89"""
		self.assertAccept('These are the same 007 numbers\tThe number is 7', 1.0)

	def test_sign_match(self):
		"""Signs matter."""
		self.assertAccept('The current temp is -7.5 degrees\tI lost -7.5 points', 1.0)
		self.assertReject('The current temp is -7.5 degrees\tI walked 7.5 miles', 1.0)
		self.assertReject('The difference is +7.5 degrees\tI walked 7.5 miles', 1.0) # questionable?
		self.assertReject('The current temp is -7.5 degrees\tI changed the value by +7.5', 1.0)

	def test_word_boundary(self):
		self.assertAccept('I am a 30something\tThat just40 should be ignored', 1.0)
	
	def test_word_boundary_dash(self):
		self.assertAccept('-30 is the number\tThe number -30', 1.0)
		self.assertAccept('The-number-30\tThe number 30', 1.0)
		self.assertReject('Beep-30\tThe number is -30', 1.0)



#!/usr/bin/env python3
from sys import stdin, stdout, stderr
from typing import Callable, List
import math
import argparse


def parse_user_args():
    """Parse the arguments necessary for this filter"""
    parser = argparse.ArgumentParser(description="Filters the lines based on the ratio between num_src_tokens and num_trg_tokens")
    parser.add_argument("--ratio-length", default=0.6, type=float)
    parser.add_argument("--filter-identical", action="store_true")
    parser.add_argument("--log", action="store_true")
    parser.add_argument("--debug", action='store_true')
    return parser.parse_args()


def compare_log(src: List[str], trg: List[str], ratio: float) -> bool:
    src_len = len(src)
    trg_len = len(trg)

    if src_len > trg_len:
        trg_len, src_len = src_len, trg_len

    ratio_len = math.log10(src_len + 1) / math.log10(trg_len + 1)

    return ratio_len >= ratio


def compare_lin(src: List[str], trg: List[str], ratio: float) -> bool:
    src_len = len(src)
    trg_len = len(trg)

    ratio_len = float(src_len) / float(trg_len)

    return ratio_len >= ratio and ratio_len <= 1.0 / ratio


Comparator = Callable[[List[str], List[str], float], bool]


def clean_parallel(ratio: float, filter_identical: bool, *, debug: bool=False, compare: Comparator=compare_lin) -> None:
    """Cleans the parallel dataset based on the ratio of source to target tokens and vice versa"""
    for line in stdin:
        fields = line.rstrip('\r\n').split('\t')
        if len(fields) != 2:
            stderr.write(f'SINGLE/MULTIPLE_LINES\t{line}')
            continue

        src = fields[0].strip()
        trg = fields[1].strip()

        # Remove identical lines
        if filter_identical and src.lower() == trg.lower():
            if debug:
                stderr.write(f'IDENTICAL\t{src}\t{trg}\n')
            continue

        src_toks = src.split()
        trg_toks = trg.split()

        if not compare(src_toks, trg_toks, ratio):
            if debug:
                stderr.write(f'RATIO_LENGTH: {src}\t{trg}\n')
        else:
            stdout.write(line)


if __name__ == '__main__':
    args = parse_user_args()
    clean_parallel(args.ratio_length, args.filter_identical,
        debug=args.debug,
        compare=compare_log if args.log else compare_lin)



import os
import sys

# Path to data files. Expects to find files named `$DATASET.$LANG.gz`.
DATA_PATH = os.getenv('DATA_PATH', 'data/train-parts/*.*.gz')

# Path to the file that defines the categories, and which dataset belongs to 
# which.
CATEGORIES_PATH = os.path.join(os.path.dirname(DATA_PATH), 'categories.json')

DEFAULT_CATEGORIES = [
	{'name': 'clean'},
	{'name': 'medium'},
	{'name': 'dirty'}
]

# TODO: Derive this from DATA_PATH. The `train-parts` is a mtdata compatibility
# thing. I'm now used to also have a data/clean directory there, so keeping it.
DOWNLOAD_PATH = 'data/train-parts'

# glob expression that looks for the filter files. Unfortunately you can't use
# commas and {} in this expression.
FILTER_PATH = os.pathsep.join([
	'filters/**/*.json',
	os.path.join(os.path.dirname(__file__), 'filters/**/*.json')
])

# col.py is used to apply a monolingual filter to a bilingual dataset. Needs
# to be absolute since filters can run from different cwds.
COL_PY = [sys.executable, os.path.join(os.path.dirname(__file__), 'col.py')]

# Program used to derive a sample from the dataset. Will be called like
# `./sample.py -n $SAMPLE_SIZE ...file-per-lang.gz`. Absolute because filters
# can specify their own `cwd`.
SAMPLE_PY = [sys.executable, os.path.join(os.path.dirname(__file__), 'sample.py')]

# Size of each of the three sections (head, random sample of middle, tail) of
# the dataset sample that we operate on.
SAMPLE_SIZE = int(os.getenv('SAMPLE_SIZE', '1000'))



import json
import os
import re
import threading
from contextlib import contextmanager
from enum import Enum
from glob import glob
from shlex import quote
from typing import Optional, Iterable, Union, Literal, Any, List, Dict, Iterator
from warnings import warn

import yaml
from pydantic import BaseModel, parse_obj_as, validator

from opuscleaner.config import COL_PY
from opuscleaner._util import none_throws


class FilterType(Enum):
    BILINGUAL = "bilingual"
    MONOLINGUAL = "monolingual"


class FilterParameterBase(BaseModel):
    type: str
    help: Optional[str]

    def export(self, value: Any) -> Any:
        return str(value)

    def default_factory(self) -> Any:
        return None


class FilterParameterFloat(FilterParameterBase):
    type: Literal["float"]
    min: Optional[float]
    max: Optional[float]
    default: Optional[float]

    def export(self, value: Any) -> Any:
        return float(value)


class FilterParameterInt(FilterParameterBase):
    type: Literal["int"]
    min: Optional[int]
    max: Optional[int]
    default: Optional[int]

    def export(self, value: Any) -> Any:
        return int(value)


class FilterParameterBool(FilterParameterBase):
    type: Literal["bool"]
    default: Optional[bool]

    def export(self, value: Any) -> Any:
        return bool(value)


class FilterParameterStr(FilterParameterBase):
    type: Literal["str"]
    default: Optional[str]
    allowed_values: Optional[List[str]]

    def export(self, value: Any) -> Any:
        # TODO: validate against allowed_values?
        return super().export(value)

    def default_factory(self) -> Any:
        return ''


class FilterParameterList(FilterParameterBase):
    type: Literal["list"]
    parameter: "FilterParameter"
    default: Optional[List[Any]]

    def export(self, value: Any) -> Any:
        return [
            self.parameter.export(item)
            for item in value
        ]

    def default_factory(self) -> Any:
        return []


class FilterParameterTuple(FilterParameterBase):
    type: Literal["tuple"]
    parameters: List["FilterParameter"]
    default: Optional[List[Any]]

    # TODO: Add validator that checks if len(default) == len(parameters)

    def export(self, value: Any) -> Any:
        return tuple(
            parameter.export(val)
            for parameter, val in zip(self.parameters, value)
        )

    def default_factory(self) -> Any:
        return [
            parameter.default_factory()
            for parameter in self.parameters
        ]


FilterParameter = Union[
    FilterParameterFloat,
    FilterParameterInt,
    FilterParameterBool,
    FilterParameterStr,
    FilterParameterList,
    FilterParameterTuple
]

FilterParameterList.update_forward_refs()
FilterParameterTuple.update_forward_refs()


class Filter(BaseModel):
    type: FilterType
    name: str # comes from filename by default
    description: Optional[str]
    command: str
    basedir: str
    parameters: Dict[str,FilterParameter]

    @validator('parameters')
    def check_keys(cls, parameters: Dict[str,Any]) -> Dict[str,Any]:
        for var_name in parameters.keys():
            if not re.match(r"^[a-zA-Z_][a-zA-Z_0-9]*$", var_name):
                raise ValueError(f"Parameter name is not a valid bash variable: {var_name}")
        return parameters


_FILTERS: Dict[str,Filter] = {}


class FilterStep(BaseModel):
    filter: str
    parameters: Dict[str,Any]
    language: Optional[str]

    @validator('filter')
    def check_filter(cls, filter_name:str) -> str:
        global _FILTERS
        if _FILTERS and filter_name not in _FILTERS:
            raise ValueError(f'Unknown filter: `{filter_name}`')
        return filter_name

    @validator('parameters')
    def check_parameters(cls, parameters:Dict[str,Any], values:Dict[str,Any], **kwargs) -> Dict[str,Any]:
        global _FILTERS
        if _FILTERS and 'filter' in values:
            required = set(_FILTERS[values['filter']].parameters.keys())
            provided = set(parameters.keys())

            missing_keys = required - provided
            if missing_keys:
                warn(f"Missing filter parameters: {' '.join(missing_keys)}")
                # Just add their default values in that case.
                parameters |= {
                    key: parameter.default if hasattr(parameter, 'default') and parameter.default is not None else parameter.default_factory()
                    for key, parameter in _FILTERS[values['filter']].parameters.items()
                    if key in missing_keys
                }
            
            superfluous_keys = provided - required
            if superfluous_keys:
                warn(f"Provided parameters not supported by the filter: {' '.join(superfluous_keys)}")
                # Not doing anything though, might be that we have just loaded an
                # old version of the filter definition and we don't want to lose
                # any of these keys.

        return parameters

    @validator('language', always=True)
    def check_language_is_provided(cls, language:str, values:Dict[str,Any], **kwargs) -> str:
        if _FILTERS and 'filter' in values:
            if _FILTERS[values['filter']].type == FilterType.BILINGUAL and language is not None:
                raise ValueError('Cannot `language` attribute for a bilingual filter')
            elif _FILTERS[values['filter']].type == FilterType.MONOLINGUAL and language is None:
                raise ValueError('`language` attribute required for a monolingual filter')
        return language


class FilterPipeline(BaseModel):
    version: Literal[1]
    files: List[str]
    filters: List[FilterStep]


def list_filters(paths:str) -> Iterable[Filter]:
    for path in paths.split(os.pathsep):
        for filename in glob(path, recursive=True):
            try:
                with open(filename) as fh:
                    defaults = {
                        "name": os.path.splitext(os.path.basename(filename))[0],
                        "basedir": os.path.dirname(filename)
                    }
                    yield parse_obj_as(Filter, {**defaults, **json.load(fh)})
            except Exception as e:
                warn(f"Could not parse {filename}: {e}")


def set_global_filters(filters:Iterable[Filter]) -> None:
    global _FILTERS
    _FILTERS = {filter.name: filter for filter in filters}


def get_global_filters() -> Dict[str,Filter]:
    global _FILTERS
    return _FILTERS


def get_global_filter(name:str) -> Filter:
    return get_global_filters()[name]


def format_shell(val: Any) -> str:
    if isinstance(val, bool):
        return '1' if val else ''
    elif isinstance(val, tuple):
        raise NotImplementedError()
    elif isinstance(val, list):
        raise NotImplementedError()
    else:
        return str(val)


def filter_format_command(filter_definition:Filter, filter_step:FilterStep, langs:List[str], *, path_to_col:List[str]=COL_PY) -> str:
    if filter_definition.type == FilterType.BILINGUAL:
        command = filter_definition.command
    elif filter_definition.type == FilterType.MONOLINGUAL:
        columns = [langs.index(language) for language in none_throws(filter_step.language).split(',')]
        command = f'{" ".join(map(quote, path_to_col))} {",".join(map(str, columns))} {filter_definition.command}'
    else:
        raise NotImplementedError()

    if filter_definition.parameters:
        params = {
            name: props.export(filter_step.parameters[name])
            for name, props in filter_definition.parameters.items()
        }
        if 'PARAMETERS_AS_YAML' in command:
            command = f'PARAMETERS_AS_YAML={quote(yaml.safe_dump(params))}; {command}'
        else:
            vars_setter = '; '.join(f"{k}={quote(format_shell(v))}" for k, v in params.items())
            command = f'{vars_setter}; {command}'

    return command



#!/usr/bin/env python3
import sys
from subprocess import Popen, PIPE
from threading import Thread
from queue import SimpleQueue
from typing import BinaryIO, Optional, TypeVar, List


queue = SimpleQueue() # type: SimpleQueue[None|list[bytes]]

T = TypeVar("T")

def none_throws(optional: Optional[T], message: str = "Unexpected `None`") -> T:
	if optional is None:
		raise AssertionError(message)
	return optional


def parse_columns(text:str) -> List[int]:
	return sorted(int(col) for col in text.split(','))


class RaisingThread(Thread):
	"""Thread that will raise any uncaught exceptions in the thread in the
	parent once it joins again."""

	exception: Optional[Exception]

	def run(self):
		self.exception = None
		try:
			super().run()
		except Exception as exc:
			self.exception = exc

	def join(self, timeout:Optional[float]=None):
		super().join(timeout=timeout)
		if self.exception is not None:
			raise self.exception


def split(columns:List[int], queue:'SimpleQueue[None|list[bytes]]', fin:BinaryIO, fout:BinaryIO):
	try:
		field_count = None
		passthru_columns = []
		for line in fin:
			fields = line.rstrip(b'\r\n').split(b'\t')
			if field_count is None:
				field_count = len(fields)
				passthru_columns = [n for n in range(field_count) if n not in columns]
			elif field_count != len(fields):
				raise RuntimeError(f'line contains a different number of fields: {len(fields)} vs {field_count}')
			queue.put([fields[column] for column in passthru_columns])
			for column in columns:
				fout.write(fields[column] + b'\n')
	except BrokenPipeError:
		pass
	finally:
		try:
			fout.close() # might fail if BrokenPipeError
		except:
			pass
		queue.put(None) # End indicator
		fin.close()


def merge(columns:List[int], queue:'SimpleQueue[None|list[bytes]]', fin:BinaryIO, fout:BinaryIO):
	try:
		while True:
			passthru_fields = queue.get()
			if passthru_fields is None:
				if fin.readline() != b'':
					raise RuntimeError('subprocess produced more lines of output than it was given')
				break

			passthru_it = iter(passthru_fields)
			for column in range(len(passthru_fields) + len(columns)):
				if column in columns:
					field = fin.readline()
					if field == b'':
						raise RuntimeError('subprocess produced fewer lines than it was given')
					field = field.rstrip(b'\r\n')
				else:
					field = next(passthru_it)

				if column > 0:
					fout.write(b'\t')
				fout.write(field)
			fout.write(b'\n')
	except BrokenPipeError:
		pass
	finally:
		fout.close()
		fin.close()


def main():
	retval = 0

	try:
		columns = parse_columns(sys.argv[1])

		child = Popen(sys.argv[2:], stdin=PIPE, stdout=PIPE)

		feeder = RaisingThread(target=split, args=[columns, queue, sys.stdin.buffer, none_throws(child).stdin])
		feeder.start()

		consumer = RaisingThread(target=merge, args=[columns, queue, none_throws(child).stdout, sys.stdout.buffer])
		consumer.start()

		retval = child.wait()
		
		if retval != 0:
			raise RuntimeError(f'subprocess exited with status code {retval}')

		feeder.join()
		consumer.join()
	except Exception as e:
		print(f'Error: {e}', file=sys.stderr)
		sys.exit(retval or 1)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
import sys
import argparse
from collections import deque, Counter
from typing import Iterable, TypeVar
from itertools import takewhile


def common_suffix(buffer: Iterable[str]) -> str:
	iters = [iter(line[::-1]) for line in buffer]
	assert len(iters) > 1
	suffix = takewhile(identical, zip(*iters))
	return "".join(t[0] for t in suffix)[::-1]


T = TypeVar('T')

def identical(elements: Iterable[T]) -> bool:
	it = iter(elements)
	first = next(it)
	return all(first == el for el in it)


def strip_suffix(lines: Iterable[str], *, minlen: int = 2, minocc: int = 5, counter: Counter=None) -> Iterable[str]:
	buffer = deque()

	suffix = ""

	for line in lines:
		if suffix and line.endswith(suffix):
			assert not buffer, "buffer should been empty"
			if counter is not None:
				counter[suffix] += 1
			yield line[:-1 * len(suffix)]

		elif suffix: # and not line ends with suffix
			assert not buffer, "buffer should been empty"
			suffix = ""
			buffer.append(line)

		else: # suffix is None
			# Make space in the buffer
			if len(buffer) == minocc:
				yield buffer.popleft()
			
			buffer.append(line)

			# If our buffer is too small to identify a suffix, don't bother
			if len(buffer) < minocc:
				continue

			# Try to identify a new common suffix
			suffix = common_suffix(buffer)

			# If the suffix is too short, it might as well be nothing
			if len(suffix) < minlen:
				suffix = ""

			# if found, empty buffer, stripping that suffix
			if suffix:
				if counter is not None:
					counter[suffix] += len(buffer)
				while buffer:
					line = buffer.popleft()
					yield line[:-1 * len(suffix)]

	# Empty buffer
	yield from buffer


if __name__  == "__main__":
	parser = argparse.ArgumentParser()
	parser.add_argument("--minlen", type=int, default=4)
	parser.add_argument("--minocc", type=int, default=5)
	parser.add_argument("--count", action="store_true")
	args = parser.parse_args()

	lines = (line.rstrip("\r\n") for line in sys.stdin)

	if args.count:
		counter = Counter()
	else:
		counter = None

	for line in strip_suffix(lines, minlen=args.minlen, minocc=args.minocc, counter=counter):
		print(line, file=sys.stdout)

	if counter:
		for suffix, count in counter.most_common():
			print(f"{suffix}\t{count}", file=sys.stderr)



#!/usr/bin/env python3
import sys
import re

def fix(text:str)->str:
	return re.sub(r'^[\'‘"“„](.+?)["”;]*$', r'\1', text)

for line in sys.stdin:
	fields = line.rstrip("\r\n").split("\t")

	fields = [fix(field).strip() for field in fields]

	if all(len(field) > 0 for field in fields):
		print("\t".join(fields))



import os
import json
from fastapi import FastAPI
from pydantic import BaseModel, parse_obj_as, validator
from typing import Any, List, Dict, TextIO, cast

from opuscleaner.config import CATEGORIES_PATH, DATA_PATH, DEFAULT_CATEGORIES


class Category(BaseModel):
	name: str

	@validator('name')
	def name_must_not_be_empty(cls, value:str) -> str:
		assert len(value.strip()) > 0, 'must not be empty'
		return value.strip()


class CategoryMapping(BaseModel):
	categories: List[Category]
	mapping: Dict[str,List[str]]

	@validator('categories')
	def categories_must_be_unique(cls, value:List[Category]) -> List[Category]:
		assert len(set(category.name.strip() for category in value)) == len(value), 'categories must have unique names'
		return value

	@validator('mapping')
	def mapping_must_only_contain_categories(cls, value:Dict[str,List[str]], values: Dict[str,Any], **kwargs) -> Dict[str,List[str]]:
		assert len(set(value.keys()) - set(category.name.strip() for category in values.get('categories', ''))) == 0, 'mapping must only contain keys that are defined in `categories`'
		return value


def read_categories(fh:TextIO) -> CategoryMapping:
	return parse_obj_as(CategoryMapping, json.load(fh))

def write_categories(mapping:CategoryMapping, fh:TextIO) -> None:
	json.dump(mapping.dict(), fh, indent=2)


app = FastAPI()

@app.get('/')
def get_mapping() -> CategoryMapping:
	if os.path.exists(CATEGORIES_PATH):
		with open(CATEGORIES_PATH, 'r') as fh:
			return read_categories(fh)
	else:
		return CategoryMapping(categories=DEFAULT_CATEGORIES, mapping=dict())

@app.put('/')
def update_categories(body:CategoryMapping) -> None:
	with open(CATEGORIES_PATH, 'w') as fh:
		write_categories(body, fh)



import time
from collections import deque
from functools import wraps
from itertools import count
from json import JSONEncoder
from threading import Thread, get_ident, main_thread
from queue import SimpleQueue
from typing import Callable, Protocol, Iterator, Optional, IO, Type, TypeVar
from uuid import uuid1, UUID


def iter_queue(queue):
	while True:
		task = queue.get()
		if task is None:
			break
		yield task


class Span:
	"""Log the duration (enter and exit) of a set of instructions"""

	def __init__(self, logger:'Logger', name:str, extra:dict=dict()):
		self.logger = logger
		self.name = name
		self.extra = extra
		self.span = None

	def __enter__(self) -> 'Span':
		self.span = self.logger.push(self.name, **self.extra, type='span', start=time.monotonic_ns())
		return self

	def __exit__(self, typ, value, traceback):
		assert self.span is not None
		span = self.logger.pop()
		assert self.span == span
		self.logger.update(self.span, end=time.monotonic_ns(), error=repr(value) if value is not None else None)

	def event(self, name, **kwargs) -> UUID:
		return self.logger.event(name, type='event', parent=self.span, **kwargs)


class Handler(Protocol):
	"""Handler receives log records to handle. For example, by writing them as
	JSON to a file, or putting them on a queue to be processed in another thread.
	"""
	def emit(self, record:dict) -> None:
		pass


class FallbackJSONEncoder(JSONEncoder):
	"""JSONEncoder that just calls `str(obj)` in case it has no built-in
	conversion for the type."""
	def default(self, obj):
		return str(obj)


class NullHandler(Handler):
	"""Handler that does nothing, like printing to /dev/null."""
	def emit(self, record:dict) -> None:
		pass


class StreamHandler(Handler):
	"""Writes log records as JSON lines to a file."""
	def __init__(self, stream:IO[str]):
		self.stream = stream
		self.encoder = FallbackJSONEncoder()

	def emit(self, record:dict) -> None:
		for chunk in self.encoder.iterencode(record):
			self.stream.write(chunk)
		self.stream.write('\n')


def _queue_to_handler(queue: SimpleQueue, handler: Handler):
	for record in iter_queue(queue):
		handler.emit(record)


class ThreadEmitter(Handler):
	"""Handler that puts log records onto a queue"""
	def __init__(self, queue):
		self.queue = queue

	def emit(self, record:dict):
		self.queue.put(record)


class ThreadReceiver:
	"""Context manager that will run a thread in the background to capture log
	records emitted by ThreadEmitter handlers, and forward them to a single
	handler. Make these emitters through `make_hander()`.
	Note that the handler is run in another thread, so if your handler is writing
	to say stderr and you're also writing to stderr on the main thread, you will
	need to do some coordination through locking.
	"""
	def __init__(self, handler:Handler):
		self.handler = handler
		self.queue = SimpleQueue()

	def __enter__(self):
		self.thread = Thread(target=_queue_to_handler, args=[self.queue, self.handler])
		self.thread.start()
		return self

	def __exit__(self, typ, value, traceback):
		self.queue.put(None)
		self.thread.join()

	def make_handler(self) -> Handler:
		return ThreadEmitter(self.queue)


class Logger:
	"""Logger object that tracks the stack when using spans."""
	handler: Handler
	serial: Iterator[int]
	stack: deque[UUID]

	def __init__(self, handler:Handler):
		self.handler = handler
		self.serial = count()
		self.stack = deque()

	def span(self, name:str, **kwargs) -> Span:
		"""Start recording a span. Use as context `with logger.span('name'):`"""
		return Span(self, name, kwargs)

	def event(self, name:str, **kwargs) -> UUID:
		"""Record a singular event. If inside the context of a span, the event will
		be associated with that span."""
		event_id = uuid1(get_ident(), next(self.serial))
		self.handler.emit({
			'id': event_id,
			'parent': self.stack[-1] if len(self.stack) > 0 else None,
			'name': name,
			**kwargs,
		})
		return event_id

	def update(self, event_id:UUID, **kwargs) -> None:
		"""Update a particular event, e.g. to add more context."""
		self.handler.emit({
			'id': event_id,
			**kwargs
		})

	def push(self, name:str, **kwargs) -> UUID:
		"""Emit an event and put it onto the stack. Same a `span().__enter__()`, it
		is better to use `with span():` in most cases."""
		event_id = self.event(name, **kwargs)
		self.stack.append(event_id)
		return event_id

	def pop(self) -> UUID:
		"""Pops event of the stack."""
		return self.stack.pop()


_main_thread_id = main_thread().ident

_context = None


class Context:
	"""Logging context. This deals with having multiple loggers on multiple 
	threads all combine into the same event stream.
	Generally you'd have something like `with logger.Context() as ctx: main()`
	in your app. You can access the current context's logger through
	`logger.get_logger()` as well as `ctx.get_logger()`.
	"""
	def __init__(self, *, file:Optional[IO[str]]=None):
		if file:
			self.handler = StreamHandler(file)
		else:
			self.handler = NullHandler()

		self.receiver = ThreadReceiver(self.handler)

		self.loggers = {
			_main_thread_id: Logger(self.handler)
		}


	def get_logger(self):
		thread_id = get_ident()

		if thread_id not in self.loggers:
			self.loggers[thread_id] = Logger(self.receiver.make_handler())
			self.loggers[thread_id].stack = deque(self.loggers[_main_thread_id].stack) # TODO what about threads starting threads?

		return self.loggers[thread_id]

	def __enter__(self):
		global _context
		assert _context is None
		self.receiver.__enter__()
		_context = self
		return self

	def __exit__(self, typ, value, traceback):
		global _context
		assert _context is self
		self.receiver.__exit__(typ, value, traceback)
		_context = None


def get_logger() -> Logger:
	"""Shortcut for Context().get_logger()"""
	if _context is None:
		raise RuntimeError('called get_logger() outside logging context')
	return _context.get_logger()


def event(name:str, **kwargs) -> UUID:
	"""Shortcut for get_logger().event()"""
	return get_logger().event(name, **kwargs)


def update(**kwargs):
	"""Shortcut for get_logger().update(current_span, ...)"""
	logger = get_logger()
	event_id = logger.stack[-1]
	logger.update(event_id, **kwargs)


def span(name:str, **kwargs) -> Span:
	"""Shortcut for get_logger().span()"""
	return get_logger().span(name, **kwargs)


# TODO: once Python3.11:
# P = ParamSpec('P')
# R = TypeVar('R')
# def trace(fn:Callable[P,R]) -> Callable[P,R]

T = TypeVar('T', bound=Callable)

def trace(fn:T) -> T:
	"""Decorator for wrapping each call to this function with
	```
	with get_logger().span(__name__):
    fn()
	```
	"""
	@wraps(fn)
	def wrapper(*args, **kwargs):
		with get_logger().span(fn.__name__):
			return fn(*args, **kwargs)
	return wrapper # type:ignore


T = TypeVar('T')

def trace_context(cls:Type[T]) -> Type[T]:
	"""Similar to `@trace`, but for a class with __enter__ and __exit__."""
	class Wrapper(cls):
		__span: Span

		def __enter__(self):
			self.__span = get_logger().span(cls.__name__).__enter__()
			return super().__enter__()

		def __exit__(self, typ, value, traceback):
			# add an __exit__ event to make it possible to measure how long the
			# wrapped __exit__ actually takes.
			self.__span.event('__exit__')
			try:
				super().__exit__(typ, value, traceback)
			finally:
				self.__span.__exit__(typ, value, traceback)

	return Wrapper



#!/usr/bin/env python3
from sys import stdin, stdout, stderr
from typing import Optional
import argparse
import re
from clean_common import CHARS

def parse_user_args():
    """Parse the arguments necessary for this filter"""
    parser = argparse.ArgumentParser(description="Filters the lines based on the ratio between alphabetic characters in a line from the language and others")
    parser.add_argument("--ratio-words-src", default=0.6, type=float, help='Ratio between words and non words (eg numbers, foreign words) in a src sentence.')
    parser.add_argument("--ratio-words-trg", default=0.6, type=float, help='Ratio between words and non words (eg numbers, foreign words) in a trg sentence.')
    parser.add_argument("--ratio-alpha-src", default=0.4, type=float, help='Ratio between characters from the src language compared to all characters (eg numbers, emoji, punctuation, etc...)')
    parser.add_argument("--ratio-alpha-trg", default=0.4, type=float, help='Ratio between characters from the trg language compared to all characters (eg numbers, emoji, punctuation, etc...)')
    parser.add_argument("--src-lang", default="en", type=str, choices=list(CHARS.keys()))
    parser.add_argument("--trg-lang", type=str, choices=list(CHARS.keys()))
    parser.add_argument("--debug", action='store_true')
    return parser.parse_args()

def clean_parallel(src_lang: str, ratio_words_src: float, ratio_alpha_src: float,\
trg_lang: Optional[str], ratio_words_trg: float, ratio_alpha_trg: float,\
 debug: bool = True) -> None:
    """Cleans the parallel (or monolingual) dataset based on the number of characters"""
    for line in stdin:
        fields = line.rstrip('\r\n').split('\t')
        if len(fields) == 1:
            src = fields[-1].strip()
            trg = None
        else: # Assumes that the multiline filter already run
            src = fields[-2].strip()
            trg = fields[-1].strip()

        if src_lang in CHARS:
            src_toks = src.split()
            src_len = len(src_toks)
            if src_len==0:
                if debug:
                    stderr.write(f'EMPTY_SRC\t{src}\t{trg}\n')
                continue

            num_words = sum(
                [1 if re.match(CHARS[src_lang], t, re.IGNORECASE) else 0 for t in src_toks])
            if num_words / float(src_len) < ratio_words_src:
                if debug:
                    stderr.write(f'RATIO_WORDS_SRC\t{src}\t{trg}\n')
                continue

            char_alpha = len(re.findall(CHARS[src_lang], src, re.IGNORECASE))
            if char_alpha / float(len(src.replace(' ', ''))) < ratio_alpha_src:
                if debug:
                    stderr.write(f'RATIO_ALPHA_SRC\t{src}\t{trg}\n')
                continue

        if trg is not None and trg_lang in CHARS:
            trg_toks = trg.split()
            trg_len = len(trg_toks)
            if trg_len==0:
                if debug:
                    stderr.write(f'EMPTY_TRG\t{src}\t{trg}\n')
                continue

            num_words = sum(
                [1 if re.match(CHARS[trg_lang], t, re.IGNORECASE) else 0 for t in trg_toks])
            if num_words / float(trg_len) < ratio_words_trg:
                if debug:
                    stderr.write(f'RATIO_WORDS_TRG\t{src}\t{trg}\n')
                continue

            char_alpha = len(re.findall(CHARS[trg_lang], trg, re.IGNORECASE))
            if char_alpha / float(len(trg.replace(' ', ''))) < ratio_alpha_trg:
                if debug:
                    stderr.write(f'RATIO_ALPHA_TRG\t{src}\t{trg}\n')
                continue
        # If none of our filters have failed, we're good to go
        stdout.write(line)


if __name__ == '__main__':
    args = parse_user_args()
    clean_parallel(src_lang=args.src_lang, ratio_words_src=args.ratio_words_src, ratio_alpha_src=args.ratio_alpha_src,\
        trg_lang=args.trg_lang, ratio_words_trg=args.ratio_words_trg, ratio_alpha_trg=args.ratio_alpha_trg, debug=args.debug)



#!/usr/bin/env python3
import argparse
import random
import subprocess
import sys

from contextlib import ExitStack, contextmanager
from itertools import count
from math import exp, log, floor
from typing import IO, Iterator
from typing import TypeVar, Iterable, Iterator, Generic, List, Tuple, IO


T = TypeVar('T')


def reservoir_sample(k:int, it:Iterable[T], *, rand:random.Random = random._inst, sort:bool=False) -> List[T]:
	"""Take k samples from iterable by reading from start to end. If sort is
	True, it will return the selected samples in the order they appeared in.
	"""
	sample: List[Tuple[int,T]] = []

	numbered_it = enumerate(it)

	i = 0

	for i, (_, line) in zip(range(k), numbered_it):
		sample.append((i, line))

	w = exp(log(rand.random())/k)

	try:
		while True:
				next_i = i + floor(log(rand.random()) / log(1 - w)) + 1

				# Skip forward
				while i < next_i:
					i, line = next(numbered_it)

				sample[rand.randrange(k)] = (i, line) # type:ignore
				w = w * exp(log(rand.random()) / k)
	except StopIteration:
		pass

	if sort:
		return [line for _, line in sorted(sample)]
	else:
		return [line for _, line in sample]


class Tailer(Iterable[T]):
	"""Functions as an iterator that returns all but the last K lines. Those lines
	you can read from `tail`."""

	def __init__(self, k:int, it:Iterable[T]):
		self.sample: List[T] = [] # ring buffer of (at maximum) length k
		self.k = k
		self.i = 0
		self.it = iter(it)

	def __iter__(self) -> Iterator[T]:
		try:
			while self.i < self.k:
				self.sample.append(next(self.it))
				self.i += 1
		except StopIteration:
			# Oh less than k samples in iterable? :(
			return

		for line in self.it:
			yield self.sample[self.i % len(self.sample)]
			self.sample[self.i % len(self.sample)] = line
			self.i += 1

	@property
	def tail(self) -> List[T]:
		# In the scenario where we read less than our tail of data, we just return
		# the entire buffer in one go.
		if len(self.sample) < self.k:
			return self.sample

		return self.sample[(self.i % len(self.sample)):] + self.sample[0:(self.i % len(self.sample))]


def sample(k:int, iterable:Iterable[T], sort:bool=False) -> Iterable[Iterable[T]]:
	"""Take `k` items from the start, the end and the middle from `iterable`. If
	`sort` is True, the items in the middle will be in the order they appeared
	in."""
	it = iter(iterable)

	yield (val for _, val in zip(range(k), it))

	tailer = Tailer(k, it)

	yield reservoir_sample(k, tailer, sort=sort)

	yield tailer.tail


@contextmanager
def gunzip(path:str) -> Iterator[IO[bytes]]:
	"""Like gzip.open(), but using external gzip process which for some reason
	is a lot faster on macOS."""
	with subprocess.Popen(['gzip', '-cd', path], stdout=subprocess.PIPE) as proc:
		assert proc.stdout is not None
		yield proc.stdout

		# Context is done with proc.stdout, so we close it. It might be that it
		# isn't completely read yet, and thus proc.wait() would block otherwise.
		proc.stdout.close()
		if proc.wait() != 0:
			raise RuntimeError(f'gzip returned error code {proc.returncode} while decompressing {path}')


def magic_open_or_stdin(ctx:ExitStack, path:str) -> IO[bytes]:
	# TODO ideally we would look at the magic bytes, but that would entail
	# consuming the input file partially and then I can't pass the complete
	# file onto gzip afterwards
	if path.endswith('.gz'):
		return ctx.enter_context(gunzip(path))
	elif path == '-':
		return sys.stdin.buffer
	else:
		return ctx.enter_context(open(path, 'rb'))


def main() -> None:
	parser = argparse.ArgumentParser(description="Take a file's head, tail and a random sample from the rest.")
	parser.add_argument('-n', dest='lines', type=int, default=10, help="number of lines for each section of the sample")
	parser.add_argument('-d', dest='delimiter', type=str, default="\\t", help="column delimiter. Defaults to \\t.")
	parser.add_argument('-N', '--line-numbers', action='store_true', help="print line numbers")
	parser.add_argument('files', metavar='file', type=str, nargs='*', default=['-'], help="files to sample. Multiple files for multiple columns. Use '-' for stdin. If none, reads from stdin.")
	args = parser.parse_args()

	with ExitStack() as ctx:
		columns:List[Iterator[bytes]] = [magic_open_or_stdin(ctx, file) for file in args.files]

		if args.line_numbers:
			columns.insert(0, (str(i).encode() for i in count()))

		pairs = zip(*columns)

		delimiter = args.delimiter.replace("\\t", "\t").replace("\\n", "\n").encode()

		for section in sample(args.lines, pairs, sort=True):
			for pair in section:
				for col, entry in enumerate(pair):
					if col > 0:
						sys.stdout.buffer.write(delimiter)
					sys.stdout.buffer.write(entry.rstrip(b"\r\n"))
				sys.stdout.buffer.write(b"\n")
			sys.stdout.buffer.flush()


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
import argparse
import re
import sys
from typing import Match, TextIO


NUM_EXPR = re.compile(r"""
	(?:
		(?P<sign>(?:(?<=\s)|^)[-+])  # start with a sign if it is not attached to a word, i.e. not - in tid-30
		|\b                          # or start with a word boundary, i.e. not just30
	)
	(?:0*)                         # allow for 0 prefixes which we then ignore when comparing
	(?P<value>\d+                  # digits before the first comma or dot
		(?:[\.,:]\d+)*                # allow commas or dots, i.e. 300,000.0 but not 3,,,5
	)
	\b                             # disallow 30beep, but also 30th and 1st #TODO
""", re.X)


def normalize(numstr:Match) -> str:
	return (numstr['sign'] or '') + re.sub(r'[^\d]+', '*', numstr['value']) # ignore the decimal and digit separators


def filter_numerical_mismatch(fin: TextIO, fout: TextIO, ratio: float, *, debug: bool = False):
	for line in fin:
		cols = line.rstrip('\r').split('\t')

		assert len(cols) >= 2

		nums_left, nums_right = (set(map(normalize, re.finditer(NUM_EXPR, col))) for col in cols[:2])

		# Only bother calculating the ratio if there were any numbers to begin with
		if nums_left or nums_right:
			overlap = nums_left & nums_right
			difference = nums_left ^ nums_right

			# Big > 1.0 number if lots of overlap, small < 1.0 number if lots of differences
			line_ratio = (len(overlap) + 1) / (len(difference) + 1)

			if debug:
				print(f"{len(overlap)} / {len(difference)} : {overlap!r} | {difference!r}", file=sys.stderr)

			if line_ratio < ratio:
				continue

		fout.write(line)


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--ratio', type=float, default=1.0)
	parser.add_argument('--debug', action='store_true')
	args = parser.parse_args()

	filter_numerical_mismatch(sys.stdin, sys.stdout, args.ratio, debug=args.debug)



#!/usr/bin/env python3
import asyncio
import gzip
import hashlib
import json
import os
import re
import subprocess
import sys
import traceback
from contextlib import ExitStack, asynccontextmanager
from enum import Enum
from glob import glob
from itertools import chain, zip_longest
from pprint import pprint
from shutil import copyfileobj
from tempfile import TemporaryFile
from typing import NamedTuple, Optional, Iterable, TypeVar, Union, Literal, Any, AsyncIterator, Awaitable, cast, IO, List, Dict, Tuple, AsyncIterator
from warnings import warn

import yaml
from fastapi import FastAPI, HTTPException
from fastapi.encoders import jsonable_encoder
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import RedirectResponse, StreamingResponse
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel, parse_obj_as, validator, ValidationError
from starlette.datastructures import URL
from starlette.responses import FileResponse, RedirectResponse, Response
from starlette.types import Scope

from opuscleaner._util import none_throws
from opuscleaner.categories import app as categories_app
from opuscleaner.config import DATA_PATH, FILTER_PATH, COL_PY, SAMPLE_PY, SAMPLE_SIZE
from opuscleaner.datasets import list_datasets, Path
from opuscleaner.download import app as download_app
from opuscleaner.filters import filter_format_command, format_shell, get_global_filter, get_global_filters, set_global_filters, list_filters, FilterType, FilterStep, FilterPipeline
from opuscleaner.sample import sample


import mimetypes
mimetypes.add_type('application/javascript', '.js')


FRONTEND_PATH = next(iter(path
    for path in [
        os.path.join(os.path.dirname(__file__), 'frontend'),
        os.path.join(os.path.dirname(__file__), '../frontend/dist'),
    ]
    if os.path.exists(path)
))


class Column(BaseModel):
    lang: str
    path: str
    size: int


class Dataset(BaseModel):
    name: str
    columns: List[Column]


class FilterPipelinePatch(BaseModel):
    """A list of changes to a filter pipeline (used when updating filters)"""
    filters: List[FilterStep]


def dataset_path(name:str, template:str) -> str:
    # TODO: fix this hack to get the file path from the name this is silly we
    # should just use get_dataset(name).path or something
    root = DATA_PATH.split('*')[0]

    # If the dataset name is a subdirectory, do some hacky shit to get to a
    # .sample.gz file in said subdirectory.
    parts = name.rsplit('/', maxsplit=2)
    if len(parts) == 2:
        root = os.path.join(root, parts[0])
        filename = parts[1]
    else:
        filename = parts[0]

    return os.path.join(root, template.format(filename))


def sample_path(name:str, langs:Iterable[str]) -> str:
    languages = '.'.join(sorted(langs))
    return dataset_path(name, f'.sample.{{}}.{languages}')


def filter_configuration_path(name:str) -> str:
    return dataset_path(name, '{}.filters.json')


async def compute_sample(name:str, columns:List[Tuple[str,Path]]) -> None:
    langs = [lang for lang, _ in columns]
    with TemporaryFile() as tempfile:
        proc = await asyncio.subprocess.create_subprocess_exec(
            *SAMPLE_PY,
            '-n', str(SAMPLE_SIZE),
            *[str(file.resolve()) for _, file in columns],
            stdout=tempfile,
            stderr=asyncio.subprocess.PIPE)

        _, stderr = await proc.communicate()

        if proc.returncode != 0:
            raise Exception(f'sample.py returned {proc.returncode}: {stderr.decode()}')

        tempfile.seek(0)

        with open(sample_path(name, langs), 'wb') as fdest:
            copyfileobj(tempfile, fdest)


class FilterOutput(NamedTuple):
    langs: List[str] # order of columns
    returncode: int
    stdout: bytes
    stderr: bytes


class ParsedFilterOutput(BaseModel):
    """JSON serializable version of FilterOutput that has stdout parsed into
       an array of dicts, with a field per language.
    """
    returncode: int
    stdout: List[Dict[str,str]]
    stderr: str

    def __init__(self, output:FilterOutput):
        lines = []

        for lineno, line in enumerate(output.stdout.rstrip(b'\r\n').split(b'\n'), start=1):
            values = []
            for colno, field in enumerate(line.rstrip(b'\r').split(b'\t'), start=1):
                try:
                    values.append(field.decode())
                except UnicodeDecodeError as e:
                    values.append(f'[Error: Cannot decode line {lineno} column {colno}: {e!s}]')
            lines.append(dict(zip_longest(output.langs, values, fillvalue='')))

        super().__init__(
            returncode=output.returncode,
            stdout=lines,
            stderr=output.stderr.decode())


class SampleCacheEntry(NamedTuple):
    checksum: bytes
    future: asyncio.Task#[FilterOutput]


sample_cache: Dict[str,List[SampleCacheEntry]] = {}


def cache_hash(obj: Any, seed: bytes = bytes()) -> bytes:
    impl = hashlib.sha256(seed)
    impl.update(json.dumps(obj, sort_keys=True).encode())
    return impl.digest()


async def get_dataset_sample(name:str, columns:List[Tuple[str,Path]]) -> FilterOutput:
    langs = [lang for lang, _ in columns]

    if not os.path.exists(sample_path(name, langs)):
        await compute_sample(name, columns)

    with open(sample_path(name, langs), 'rb') as fh:
        stdout = fh.read()

    return FilterOutput([lang for lang, _ in columns], 0, stdout, bytes())


async def exec_filter_step(filter_step: FilterStep, langs: List[str], input: bytes) -> Tuple[bytes,bytes]:
    filter_definition = get_global_filter(filter_step.filter)

    command = filter_format_command(filter_definition, filter_step, langs)

    # Make sure the path to the python binary (and the installed utils)
    # is in the PATH variable. If you load a virtualenv this happens by
    # default, but if you call it with the virtualenv's python binary
    # directly it wont.
    pyenv_bin_path = os.path.dirname(sys.executable)
    os_env_bin_paths = os.environ.get('PATH', '').split(os.pathsep)
    filter_env = {
        **os.environ,
        'PATH': os.pathsep.join([pyenv_bin_path] + os_env_bin_paths)
    } if pyenv_bin_path not in os_env_bin_paths else None

    p_filter = await asyncio.create_subprocess_shell(command,
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=filter_definition.basedir,
        env=filter_env)

    # Check exit codes, testing most obvious problems first.
    stdout, stderr = await p_filter.communicate(input=input)

    return FilterOutput(langs, p_filter.returncode, stdout, stderr)


def cancel_cached_tasks(name:str, offset:int) -> None:
    assert offset > 0 # offset == 0 is sample.py itself, and should never be cancelled through this method.
    for entry in sample_cache[name][offset:]:
        entry.future.cancel()
    del sample_cache[name][offset:]


async def get_sample(name:str, filters:List[FilterStep]) -> AsyncIterator[FilterOutput]:
    columns: List[Tuple[str,Path]] = list_datasets(DATA_PATH)[name]
    langs = [lang for lang, _ in columns]

    checksum = cache_hash([
        (name, str(path), path.stat().st_mtime)
        for name, path in columns
    ])

    # If we don't have a sample stored, generate one. Doing it in bytes because
    # it might save us parsing utf-8 (also assumptions! It it utf-8?)
    if not name in sample_cache or sample_cache[name][0].checksum != checksum:
        # If the there is a sampler running, but it is outdated, cancel it.
        if name in sample_cache:
            sample_cache[name][0].future.cancel()

        sample_cache[name] = [
            SampleCacheEntry(
                checksum=checksum,
                future=asyncio.create_task(get_dataset_sample(name, columns))
            )
        ]

    # Using `asyncio.shield()` so that when get_sample gets cancelled because
    # the request got aborted, we don't stop computing the sample.
    sample = await asyncio.shield(sample_cache[name][0].future)

    # Return a clean unfiltered sample first
    yield sample

    for i, filter_step in enumerate(filters, start=1):
        filter_definition = get_global_filter(filter_step.filter)

        # Cache invalidation checksum, includes:
        # - arguments
        # - command itself
        # - stdin input (via checksum of previous step)
        checksum = cache_hash(
            jsonable_encoder(filter_step),
            cache_hash(jsonable_encoder(filter_definition),
                sample_cache[name][i-1].checksum))

        # If we do not have a cache entry for this point
        if len(sample_cache[name]) <= i or sample_cache[name][i].checksum != checksum:
            # Invalidate all the cache after this step
            cancel_cached_tasks(name, i)

            sample_cache[name].append(SampleCacheEntry(
                checksum=checksum,
                future=asyncio.create_task(exec_filter_step(filter_step, langs, sample.stdout))
            ))

            assert len(sample_cache[name]) == i + 1

        # Again shield from cancellation. If we don't need this filter's output
        # in the next `get_sample()`, `cancel_cached_tasks()` will cancel it.
        sample = await asyncio.shield(sample_cache[name][i].future)

        # Return the (partially) filtered sample
        yield sample


    # if there are additional steps left in the cache, remove them
    if len(sample_cache[name]) > len(filters) + 1:
        cancel_cached_tasks(name, len(filters) + 1)


def stream_jsonl(iterable:AsyncIterator[Any]) -> StreamingResponse:
    return StreamingResponse(
        (
            json.dumps(jsonable_encoder(line), separators=(',', ':')).encode() + b"\n"
            async for line in iterable
        ),
        media_type='application/json')


@asynccontextmanager
async def lifespan(app: FastAPI):
    """
    Place to do the start-up and shut-down operations of the web service.
    See https://fastapi.tiangolo.com/advanced/events/#lifespan-events
    """
    set_global_filters(list_filters(FILTER_PATH))
    yield
    set_global_filters([])


app = FastAPI(lifespan=lifespan)

app.add_middleware(GZipMiddleware, minimum_size=512)

@app.get('/api/datasets/')
def api_list_datasets() -> List[Dataset]:
    return [
        Dataset(name=name, columns=[
            Column(lang=lang, path=file.name, size=file.stat().st_size)
            for lang, file in columns
        ])
        for name, columns in list_datasets(DATA_PATH).items()
    ]


@app.get('/api/datasets/{name:path}/')
def api_get_dataset(name:str) -> Dataset:
    columns = list_datasets(DATA_PATH).get(name)

    if not columns:
        raise HTTPException(status_code=404, detail='Dataset not found')

    return Dataset(name=name, columns=[
        Column(lang=lang, path=file.name, size=file.stat().st_size)
        for lang, file in columns
    ])


@app.get('/api/datasets/{name:path}/sample')
def api_get_sample(name:str) -> Response:
    return stream_jsonl(ParsedFilterOutput(output) async for output in get_sample(name, []))


@app.post('/api/datasets/{name:path}/sample')
def api_get_filtered_sample(name:str, filters:List[FilterStep]) -> Response:
    return stream_jsonl(ParsedFilterOutput(output) async for output in get_sample(name, filters))


def make_pipeline(name:str, filters:List[FilterStep] = []) -> FilterPipeline:
    columns = list_datasets(DATA_PATH)[name]
    return FilterPipeline(
        version=1,
        files=[file.name for _, file in columns],
        filters=filters
    )


@app.get('/api/datasets/{name:path}/configuration.json')
def api_get_dataset_filters(name:str) -> FilterPipeline:

    if not os.path.exists(filter_configuration_path(name)):
        return make_pipeline(name)

    with open(filter_configuration_path(name), 'r') as fh:
        data = json.load(fh)
        try:
            return parse_obj_as(FilterPipeline, data)
        except ValidationError:
            try:
                # Backwards compatibility
                return make_pipeline(name, parse_obj_as(List[FilterStep], data))
            except ValidationError:
                # Last resort case
                return make_pipeline(name)



@app.patch('/api/datasets/{name:path}/configuration.json')
def api_update_dataset_filters(name:str, patch:FilterPipelinePatch):
    pipeline = make_pipeline(name, patch.filters)
    with open(filter_configuration_path(name), 'w') as fh:
        return json.dump(pipeline.dict(), fh, indent=2)


@app.get('/api/datasets/{name:path}/configuration-for-opusfilter.yaml')
def api_get_dataset_filters_as_openfilter(name:str) -> Response:
    if not os.path.exists(filter_configuration_path(name)):
        raise HTTPException(status_code=404, detail='Dataset not found')

    with open(filter_configuration_path(name), 'r') as fh:
        data = json.load(fh)

    pipeline = parse_obj_as(FilterPipeline, data)

    opusfilter_config: Dict[str,Any] = {
        'steps': []
    }

    input_files = pipeline.files

    preprocess_steps = []

    filter_steps: List[Dict[str,Any]] = []

    for step in pipeline.filters:
        if (match := re.search(r'\bopusfilter\.preprocessors\.(\w+)\b', get_global_filter(step.filter).command)):
            preprocess_steps.append({
                str(match.group(1)): step.parameters
            })
        elif (match := re.search(r'\bopusfilter\.filters\.(\w+)\b', get_global_filter(step.filter).command)):
            filter_steps.append({
                str(match.group(1)): step.parameters
            })
        elif get_global_filter(step.filter).type == FilterType.BILINGUAL:
            filter_steps.append({
                'OpusCleanerFilter': {
                    'filter': step.filter,
                    'parameters': step.parameters
                },
                'module': 'opuscleaner.opusfilter_compat'
            })
        elif get_global_filter(step.filter).type == FilterType.MONOLINGUAL:
            filter_steps.append({
                'OpusCleanerFilter': {
                    'filter': step.filter,
                    'parameters': step.parameters
                },
                'module': 'opuscleaner.opusfilter_compat'
            })
        else:
            raise ValueError(f'Cannot convert "{step.filter}" to opusfilter configuration')

    if preprocess_steps:
        output_files = [
            os.path.join(os.path.dirname(file), 'preprocessed.' + os.path.basename(file))
            for file in pipeline.files
        ]

        opusfilter_config['steps'].append({
            'type': 'preprocess',
            'parameters': {
                'inputs': input_files,
                'outputs': output_files,
                'preprocessors': preprocess_steps
            }
        })

        input_files = output_files

    if filter_steps:
        output_files = [
            os.path.join(os.path.dirname(file), 'filtered.' + os.path.basename(file))
            for file in pipeline.files
        ]

        opusfilter_config['steps'].append({
            'type': 'filter',
            'parameters': {
                'inputs': input_files,
                'outputs': output_files,
                'filters': filter_steps
            }
        })

        input_files = output_files

    return Response(yaml.safe_dump(opusfilter_config, sort_keys=False), media_type='application/yaml')


@app.get('/api/filters/')
def api_get_filters():
    set_global_filters(list_filters(FILTER_PATH))
    return get_global_filters()


@app.get('/')
def redirect_to_interface():
    return RedirectResponse('/frontend/index.html')


app.mount('/frontend/', StaticFiles(directory=FRONTEND_PATH, html=True), name='static')

app.mount('/api/download/', download_app)

app.mount('/api/categories/', categories_app)

def main_serve(args):
    import uvicorn
    uvicorn.run(f'opuscleaner.server:app', host=args.host, port=args.port, reload=args.reload, log_level='info')


async def sample_all_datasets(args):
    tasks = []

    for name, columns in list_datasets(DATA_PATH).items():
        langs = [lang for lang, _ in columns]
        if not os.path.exists(sample_path(name, langs)):
            print(f"Sampling {name}...", file=sys.stderr)
            tasks.append([name, columns])

    for task, result in zip(tasks, await asyncio.gather(*[compute_sample(*task) for task in tasks], return_exceptions=True)):
        if isinstance(result, Exception):
            print(f"Could not compute sample for {task[0]}: {result!s}", file=sys.stderr)


def main_sample(args):
    asyncio.run(sample_all_datasets(args))


def main_list_commands(args):
    print("Error: No command specified.\n\n"
          "Available commands:\n"
          "  serve      run webserver\n"
          "  sample     sample all datasets\n"
          "", file=sys.stderr)
    sys.exit(1)


def main(argv=sys.argv):
    import argparse

    parser = argparse.ArgumentParser(description='Fill up those seats on your empty train.')
    parser.set_defaults(func=main_list_commands)
    subparsers = parser.add_subparsers()

    parser_serve = subparsers.add_parser('serve')
    parser_serve.add_argument('--host', type=str, default='127.0.0.1', help='Bind socket to this host. (default: 127.0.0.1)')
    parser_serve.add_argument('-p', '--port', type=int, default=8000, help='Bind socket to this port. (default: 8000)')
    parser_serve.add_argument('--reload', action='store_true', help='Enable auto-reload.')
    parser_serve.set_defaults(func=main_serve)

    parser_sample = subparsers.add_parser('sample')
    parser_sample.set_defaults(func=main_sample)

    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Lists datasets given a directory. It works by scanning the directory and looking for gz files."""
from glob import glob
from itertools import groupby
from pathlib import Path as Path
from typing import Dict, List, Tuple

from opuscleaner.config import DATA_PATH


def list_datasets(path:str) -> Dict[str,List[Tuple[str,Path]]]:
    """Lists datasets given a directory. Scans the directories and returns a dictionary of the
    datasets encoutered. Dictionary looks like {dataset_name : { lang: path}}"""
    root = Path(path.split('*')[0])

    entries = (Path(entry) for entry in glob(path, recursive=True))

    files = [
        entry
        for entry in entries
        if entry.is_file()
        and entry.name.endswith('.gz')
        and not entry.name.startswith('.')
    ]

    datasets = [
        (name, list(files))
        for name, files in groupby(
            sorted(files, key=lambda entry: str(entry)),
            key=lambda entry: str(entry.relative_to(root)).rsplit('.', 2)[0])
    ]

    return {
        name: [
            (entry.name.rsplit('.', 2)[1], entry)
            for entry in files
        ]
        for name, files in datasets
    }


def main() -> None:
    import sys
    import pprint
    if len(sys.argv) == 1:
        pprint.pprint(list_datasets(DATA_PATH))
    else:
        pprint.pprint(list_datasets(sys.argv[1]))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Normalize whitespace filter.

Strips all leading and trailing whitespaces. Optionally, collapses all groups
of whitespaces into a single space

"""
import argparse
import sys


def collapse_whitespace(s):
    """Collapses whitespace groups into a single space."""
    return " ".join(s.split())


def clean(collapse):
    """Runs the filter."""

    for line in sys.stdin:
        line = line.strip()

        if collapse:
            line = collapse_whitespace(line)

        sys.stdout.write(f"{line}\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--collapse", action="store_true",
        help="Collapse whitespace groups into single spaces")
    args = parser.parse_args()
    clean(args.collapse)



#!/usr/bin/env python3
import sys
import re
import argparse


mapping = [
	("\u0083", "É"),	# D<U+0083>CLARATION
	("\u0088", "à"),	# permettre <U+0088> lŐUnion
	("\u0089", "â"),	# probants obtenus gr<U+0089>ce aux
	("\u008d", "ç"),	# Recettes per<U+008D>ues
	("\u008e", "é"),	# d<U+008E>claration
	("\u008f", "è"),	# ci-apr<U+008F>s
	("\u0090", "ê"),	# doivent <U+0090>tre exclus
	("\u0091", "ë"),	# lŐexercice ont d<U+009E> être
	("\u0094", "î"),	# L'Agence reconna<U+0094>t
	("\u0099", "ô"),	# xLes contr<U+0099>les
	("\u009d", "ù"),	# dans la mesure o<U+009D> il
	("\u009e", "û"),	# l'exercice ont d<U+009E> être reportés
	("\u0092", "í"),    # V<U+0092>tor Manuel da SILVA CALDEIRA
	("Ő", "'"),         # lŐexercice
	("ă", "'"),         # ăAutoriteitÓ
	("Ó", "'"),
	("􏳕", "ë"),        # financi􏳕le
	("¬ ", ""),
]


class Translator:
	def __init__(self, mapping):
		self.mapping = {entry[0]: entry[1] for entry in mapping}
		self.pattern = re.compile('(' + '|'.join(self.mapping.keys()) + ')')
		self.callback = lambda match: self.mapping[match[0]]

	def __call__(self, input):
		return re.sub(self.pattern, self.callback, input)


def parse_user_args():
    parser = argparse.ArgumentParser(description="Fixes select encoding issues on the French side of the ELITR ECA dataset.")
    return parser.parse_args()


if __name__ == "__main__":
	args = parse_user_args()
	for line in sys.stdin:
		sys.stdout.write(Translator(mapping)(line))




#!/usr/bin/env python3
from sys import stdin
import spacy_pkuseg as pkuseg

seg = pkuseg.pkuseg() #load the default model
for line in stdin:
    text = seg.cut(line.strip())
    print(" ".join(text))



#!/usr/bin/env python3
"""Stand-alone filter pipeline runner. Executes all of the filters defined in
a dataset filtering pipeline created by empty-train in their own process and
links them together through pipes. Can read from stdin but by default reads
the dataset from the same folder as the pipeline configuration file.
"""
import argparse
import json
import os
import re
import signal
import sys
import traceback
from queue import SimpleQueue
from shlex import quote
from shutil import copyfileobj
from subprocess import Popen, PIPE
from tempfile import NamedTemporaryFile, TemporaryDirectory
from threading import Thread
from typing import Dict, List, IO, Optional, TypeVar, Iterable, Tuple, NamedTuple, Union
from io import TextIOWrapper

from pydantic import parse_obj_as

from opuscleaner import logging
from opuscleaner.config import FILTER_PATH
from opuscleaner.filters import list_filters, set_global_filters, filter_format_command, Filter, FilterPipeline, quote, format_shell
from opuscleaner._util import none_throws, ThreadPool, CancelableQueue, Cancelled


# Queue for printing lines to stdout or stderr. None means end of input.
PrintQueue = SimpleQueue[Union[None,bytes]]

# Control queue for communicating the return code of a child back to the parent.
ControlQueue = SimpleQueue[Tuple[int,int]]

# Batches to be processed. tuple[batch index,batch path]. None means end of input.
# Using a Queue here to limit the maximum capacity.
BatchQueue = CancelableQueue[Union[None,Tuple[int,str]]]

# Batches to be merged. Same format as BatchQueue
MergeQueue = CancelableQueue[Union[None,Tuple[int,str]]]


def load_time(fh:IO[str]) -> Dict[str,float]:
    time = {}
    for line in fh:
        match = re.match(r'^(real|user|sys)\s+(\d+\.\d+)$', line.rstrip('\r\n'))
        if match:
            time[match[1]] = float(match[2])
    return time


@logging.trace
def babysit_child(n: int, child: Popen, name: str, print_queue: PrintQueue, ctrl_queue: ControlQueue, time_read_fd:Optional[int]=None) -> None:
    """Thread that looks after a child process and passes (and prefixes) all of
    its stderr to a queue. It will tell the parent thread about the end of the
    child through the ctrl_queue.
    """
    try:
        logging.update(n=n, pid=child.pid, args=child.args)

        prefix = f'[{name}] '.encode()

        for line in none_throws(child.stderr):
            print_queue.put(prefix + line)

        child.wait()

        logging.event('child_exited', retval=child.returncode)

        # If the command was wrapped by `time`, we want to read its output as
        # well. It's written to a separate pipe as to not end up in the stderr
        # of the main command.
        if time_read_fd is not None:
            with os.fdopen(time_read_fd, 'r') as fh:
                logging.update(time=load_time(fh))
    finally:
        ctrl_queue.put((n, child.returncode))


def print_lines(queue: PrintQueue, fout: IO[bytes]) -> None:
    """Thread that prints stderr lines from all the children to stderr in an
    orderly fashion.
    """
    while True:
        line = queue.get()
        if line is None:
            break
        fout.write(line)

        # Since we're writing stderr, we flush after each line to make it more
        # useful for debugging
        fout.flush()


T = TypeVar('T')

def mark_last(iterable: Iterable[T]) -> Iterable[Tuple[bool,T]]:
    it = iter(iterable)
    curr_el = next(it)
    while True:
        try:
            next_el = next(it)
            yield False, curr_el
            curr_el = next_el
        except StopIteration:
            break
    yield True, curr_el


class Child(NamedTuple):
    name: str
    process: Popen
    babysitter: Thread


@logging.trace_context
class ProcessPool:
    """Context manager for spawning and babysitting child processes that are
    siblings connected by their pipes. Exiting the context will cause it to
    block and wait for the children to finish.
    If any of the children exits early or with an error, or there was an
    uncaught exception inside the context, it will terminate all the other
    processes in the pool and raise an exception on exit. SIGPIPE errors, and
    errors caused by the pool terminating the process, are ignored.
    """
    print_prefix: str

    ctrl_queue: ControlQueue

    print_queue: PrintQueue

    environ: Dict[str,str]

    children: List[Child]

    def __init__(self, print_queue: PrintQueue, *, env:Dict[str,str]={}, print_prefix:str=''):
        self.print_prefix = print_prefix
        self.ctrl_queue = SimpleQueue()
        self.print_queue = print_queue
        self.environ = dict(env)
        self.children = []

    def start(self, name:str, cmd: Union[str,List[str]], *, shell:bool=False, time:bool=False, **kwargs) -> Popen:
        """Set up a process in the pool. Similar to Popen. `name` is used for
        identifying the process in log messages and exceptions. `time` can be
        set to True to wrap the process in `/usr/bin/time`. Furthermore all
        arguments to `Popen` are accepted.
        """
        time_read_fd, time_write_fd = None, None

        args = ([cmd] if isinstance(cmd, str) else cmd)
        
        if shell:
            args = ['/bin/sh', '-c', *args] # TODO: sorry Windows, Andriod

        # If we're measuring time, prepend `/usr/bin/time` and let it write to
        # a pipe we will read out later. Massive assumption: that pipe's buffer
        # will be sufficient for time's output.
        if time:
            time_read_fd, time_write_fd = os.pipe()
            os.set_inheritable(time_write_fd, True) # TODO is this necessary?
            args = ['/usr/bin/time', '-p', '-o', f'/dev/fd/{time_write_fd}', *args]
            kwargs['pass_fds'] = (time_write_fd, *kwargs.get('pass_fds', tuple()))
        
        child = Popen(args, **{
            **kwargs,
            'env': {
                **os.environ,
                **self.environ,
                **(kwargs.get('env') or dict())
            }
        })

        # If we have a time pipe, make sure we release our handle of the write
        # side. We just keep the read side.
        if time_write_fd:
            os.close(time_write_fd)

        n = len(self.children)
        thread = Thread(target=babysit_child, args=[n, child, name, self.print_queue, self.ctrl_queue, time_read_fd])
        thread.start()
        self.children.append(Child(name, child, thread))
        return child

    def __enter__(self) -> 'ProcessPool':
        return self

    def __exit__(self, err_type, err_inst, _) -> None:
        # Wait for the children to exit, and depending on their retval exit early
        running_children = len(self.children)

        # If we hit __exit__ due to an exception, just terminate everything
        if err_type:
            for child in self.children:
                child.process.terminate()

        # Place to store a non-zero child exit
        problem_child: Optional[Child] = None

        # Wait for all children to signal their exit
        try:
            while running_children > 0:
                child_i, retval = self.ctrl_queue.get()
                running_children -= 1

                logging.event('child_exit_received', n=child_i, retval=retval)

                # Early exit when a process errored out. SIGPIPE is retuned by
                # processes that can no longer write to the next one. E.g. when
                # `head -n 10` stops reading because it has read its 10 lines.
                if retval not in (0, -signal.SIGPIPE):
                    problem_child = self.children[child_i]
                    break
        except KeyboardInterrupt:
            # Oh user doesn't want to wait? Okay, then we terminate.
            for child in self.children:
                child.process.terminate()
            pass

        # Wait for all the processes to exit to prevent zombies
        for child in self.children:
            if child.process.returncode is None:
                child.process.wait()

        # Wait for the babysitters to exit, which happens when their process has stopped
        for child in self.children:
            child.babysitter.join()

        # If we broke out of our ctrl_queue loop we did so because there was an issue
        # with one of the children. Let's raise that to the parent.
        if not err_inst and problem_child:
            raise RuntimeError(f"Child {problem_child.name} (pid {problem_child.process.pid}) exited with {problem_child.process.returncode}")


class PipelineStep(NamedTuple):
    name: str
    command: str
    basedir: str


class Pipeline:
    """Description of a set of filter steps with all their variables filled in
    set up to execute in a certain environment. A Pipeline can either be dumped
    as a bash script, or executed on a ProcessPool.
    """
    def __init__(self, filters:Dict[str,Filter], languages: List[str], pipeline: FilterPipeline):
        self.steps: List[PipelineStep] = []

        # Make sure the path to the python binary (and the installed utils)
        # is in the PATH variable. If you load a virtualenv this happens by
        # default, but if you call it with the virtualenv's python binary 
        # directly it wont.
        pyenv_bin_path = os.path.dirname(sys.executable)
        os_env_bin_paths = os.environ.get('PATH', '').split(os.pathsep)
        self.env: Optional[Dict[str,str]] = {
            **os.environ,
            'PATH': os.pathsep.join([pyenv_bin_path] + os_env_bin_paths)
        } if pyenv_bin_path not in os_env_bin_paths else None

        # Assert we have all filters we need
        assert set(step.filter for step in pipeline.filters) - set(filters.keys()) == set()

        # Make sure the path to the python binary (and the installed utils)
        # is in the PATH variable. If you load a virtualenv this happens by
        # default, but if you call it with the virtualenv's python binary 
        # directly it wont.
        pyenv_bin_path = os.path.dirname(sys.executable)
        os_env_bin_paths = os.environ.get('PATH', '').split(os.pathsep)
        self.env: Optional[Dict[str,str]] = {
            **os.environ,
            'PATH': os.pathsep.join([pyenv_bin_path] + os_env_bin_paths)
        } if pyenv_bin_path not in os_env_bin_paths else None

        for step in pipeline.filters:
            filter_def = filters[step.filter]
            command_str = filter_format_command(filter_def, step, languages)
            self.steps.append(PipelineStep(step.filter, command_str, filter_def.basedir))

    def run(self, pool:ProcessPool, stdin:IO[bytes], stdout:IO[bytes], *, tee:bool=False, basename:str="", time:bool=False) -> None:
        """Set up all the processes on `pool`, processing `stdin` to `stdout`.
        Note that this function will return as soon as the processes have been
        set up. You will have to use the ProcessPool to wait for them to finish.
        Optionally you can `tee` the output of each filter step to a separate
        file for debugging (with the name "{basename}.step-{i}.tsv". You can 
        use `time` two wrap every filter step command in `/usr/bin/time` and
        the baby sitter will measure how much processing time the filter process
        used."""
        if not self.steps:
            copyfileobj(stdin, stdout)
            return

        for i, (is_last_step, step) in enumerate(mark_last(self.steps)):
            child = pool.start(f'{pool.print_prefix}{i}:{step.name}', step.command,
                stdin=stdin,
                stdout=stdout if is_last_step and not tee else PIPE,
                stderr=PIPE,
                cwd=step.basedir,
                env=self.env,
                shell=True,
                time=time)

            # Close our reference to the previous child, now taken over by the next child
            stdin.close()
            
            # Set stdin for next step (unless there is none, then child.stdout is None)
            if not is_last_step and not tee:
                stdin = none_throws(child.stdout)

            # If we are tee-ing for debug, shunt the output to a separate file
            # TODO: uncompressed at the moment. Might be trouble.
            if tee:
                tee_child = pool.start(f'{pool.print_prefix}{i}:tee',
                    ['tee', f'{basename}.step-{i}.tsv'],
                    stdin=stdin,
                    stdout=stdout if is_last_step else PIPE,
                    stderr=PIPE)

                stdin.close()
                stdin = none_throws(tee_child.stdout)

    def dump(self, out:IO[str]) -> None:
        """Write this pipeline as a bash script."""
        if self.env:
            for key, val in self.env:
                out.write(f'export {key}={quote(format_shell(val))}\n')

        for is_last_step, step in mark_last(self.steps):
            out.write(f'(cd {quote(format_shell(step.basedir))} && ({step.command}))')
            out.write('\n' if is_last_step else ' |\n')



def split_input(parallel: int, batch_queue: BatchQueue, batch_size:int, stdin:IO[bytes]) -> None:
    """Reads data from `stdin` and splits it into chunks of `batch_size` lines.
    These chunks are stored in temporary files, whose filenames are put onto
    `batch_queue`.
    """
    more = True

    batch_index = 0

    while more:
        fh = NamedTemporaryFile(delete=False)
        
        lines = 0

        while lines < batch_size:
            line = stdin.readline()
            if line == b'':
                more = False
                break
            fh.write(line)
            lines += 1
        
        fh.close()

        try:    
            if lines > 0:
                batch_queue.put((batch_index, fh.name))
            else:
                # Empty chunk because `len(stdin) % batch_size == 0`. No need
                # to process it further.
                os.unlink(fh.name)
        except Cancelled:
            # batch_queue got interrupted, so fn.name never made it into the
            # queue. Let's clean that up.
            os.unlink(fh.name)
            raise

        batch_index += 1

    # Tell all the runners there will be no more batches coming.
    for _ in range(parallel):
        batch_queue.put(None)


@logging.trace
def run_pipeline(print_queue:PrintQueue, batch_queue:BatchQueue, merge_queue:MergeQueue, pipeline:Pipeline, *, time:bool=False) -> None:
    """Receives an input filename from `batch_queue`, and once that has been processed
    with `pipeline`, it will post the output filename to `merge_queue`.
    stderr from any of the filter processes will be forwarded to `print_queue`.

    TODO: This could also instead run ./run.py on the input and output files
    directly as opposed to using `ProcessPool` + `pipeline.run()`.

    TODO: We can rewrite this to call `srun` on SLUM clusters so that the
    actual filtering pipeline is executed on a different node. Since input
    and output are just files on the same filesystem (depends on TMPDIR) this
    should pretty much work out of the box :O
    """
    with TemporaryDirectory() as tmpdir:
        while True:
            entry = batch_queue.get()

            # If the batcher told us they're out of batches, stop.
            if entry is None:
                break

            batch_index, filename = entry

            try:
                # Write pipeline output to tempfile that is then passed on to merger.
                stdout = NamedTemporaryFile(delete=False)

                try:
                    # Open chunk file and process pool and run the pipeline with it.
                    # The pool's __exit__() will make us wait till the pipeline is done.
                    with logging.span('run_pipeline_batch', batch_index=batch_index), \
                        open(filename, 'rb') as stdin, \
                        ProcessPool(print_queue, env={'TMPDIR': tmpdir}, print_prefix=f'{batch_index}/') as pool:
                        pipeline.run(pool, stdin, stdout, time=time)

                    stdout.close()

                    # Tell merger that they can process this batch when the time comes
                    merge_queue.put((batch_index, stdout.name))
                except Exception as exc:
                    # Didn't get to put it on the queue, delete it.
                    os.unlink(stdout.name)
                    raise
            except Exception as exc:
                # Add a bit more info, and re-raise
                raise RuntimeError(f'Error while processing batch {batch_index}') from exc
            finally:
                # Delete the input file from disk.
                os.unlink(filename)
        
        # Tell the merger that they should not be expecting more input from you.
        merge_queue.put(None)


def merge_output(parallel:int, merge_queue:MergeQueue, stdout:IO[bytes]) -> None:
    """Takes batch filenames and numbers from `merge_queue` and will concatenate
    files in the order of the batches. If batches arrive out of order, it will
    wait for the next in order batch to arrive before continuing to concatenate.
    """
    next_batch_index = 0

    pending_batches: Dict[int, str] = {}

    while True:
        # If we have the next batch, start processing it into the final output
        if next_batch_index in pending_batches:
            batch_index, filename = next_batch_index, pending_batches[next_batch_index]

            try:
                with logging.span(f'merge_output_batch', batch_index=batch_index), open(filename, 'rb') as fh:
                    copyfileobj(fh, stdout)
            except Exception as exc:
                raise RuntimeError(f'Error while merging batch {batch_index}') from exc
            finally:
                os.unlink(filename)

            next_batch_index += 1
        # If not yet, we wait on the queue to come through with (hopefully) the next batch
        elif parallel > 0:
            entry = merge_queue.get()

            # Another batch processor finished
            if entry is None:
                parallel -= 1
            else:
                batch_index, filename = entry
                assert batch_index not in pending_batches
                pending_batches[batch_index] = filename
        # next_batch_index is not in pending_batches, and there are no more batches coming from
        # any batch processors. So let's stop.
        else:
            break

    if len(pending_batches) and next_batch_index <= max(pending_batches.keys()):
        raise RuntimeError(f'Not all batches got merged: {next_batch_index=} <= {max(pending_batches.keys())=}')


@logging.trace
def run_parallel(pipeline:Pipeline, stdin:IO[bytes], stdout:IO[bytes], *, parallel:int, batch_size:int, print_queue: PrintQueue, time:bool=False) -> None:
    """Run `parallel` copies of the processing pipeline in parallel, each
    working on a batch of `batch_size` lines at a time. Batches will be cut
    from `stdin` and printed to `stdout`, in order. stderr from the filter
    processes will be forwarded to `print_queue`. `time` is forwarded to
    ProcessPool.
    """
    batch_queue: BatchQueue = CancelableQueue(parallel * 2)

    merge_queue: MergeQueue = CancelableQueue()

    with ThreadPool() as pool:
        # Splits stdin into files of `batch_size` lines, and puts those on `batch_queue`
        pool.start(split_input, parallel, batch_queue, batch_size, stdin)

        # Read `batch_queue` for batch filenames, and process them. Put output files
        # on `merge_queue`.
        for _ in range(parallel):
            pool.start(run_pipeline, print_queue, batch_queue, merge_queue, pipeline, time=time)

        # Read from `merge_queue` and combine files in order.
        pool.start(merge_output, parallel, merge_queue, stdout)

        try:
            pool.join()
        except BaseException: # Note: also catches KeyboardInterrupt
            batch_queue.cancel()
            merge_queue.cancel()
            raise


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument('--filters', '-f', type=str, default=FILTER_PATH, help='Path to directory with filter specifications')
    parser.add_argument('--input', '-i', type=argparse.FileType('rb'), help='Input tsv. If unspecified input files are read from filter json; use - to read from stdin')
    parser.add_argument('--output', '-o', type=argparse.FileType('wb'), default=sys.stdout.buffer, help='Output tsv (defaults to stdout)')
    parser.add_argument('--basedir', '-b', type=str, help='Directory to look for data files when --input is not used (defaults to same as input pipeline file)')
    parser.add_argument('--tee', action='store_true', help='Write output after each step to a separate file')
    parser.add_argument('--parallel', type=int, default=1, help='Run N parallel copies of the pipeline processing batches')
    parser.add_argument('--batch-size', type=int, default=1_000_000, help='Batch size in lines that each parallel copy processes (only if --parallel > 1)')
    parser.add_argument('--first', type=int, default=0, help='Limit reading input to the N first lines')
    parser.add_argument('--dump', action='store_true', help='Print shell script instead')
    parser.add_argument('--trace', type=argparse.FileType('a'), nargs='?', const='/dev/stderr', help='Write tracing JSON to file (defaults to stderr)')
    parser.add_argument('--time', action='store_true', help='Measure real/user/sys times for each filter step')
    parser.add_argument('pipeline', metavar='PIPELINE', type=argparse.FileType('r'), help='Pipeline steps specification file, e.g. *.filters.json')
    parser.add_argument('languages', metavar='LANG', type=str, nargs='*', help='Language codes of the columns in the input TSV. Only used when --input is set')

    args = parser.parse_args()

    with logging.Context(file=args.trace), logging.span('main'):
        # default search path for the data files is next to the configuration file
        # which is the default save location for empty-train.
        if not args.basedir:
            args.basedir = os.path.dirname(args.pipeline.name) or os.getcwd()

        if args.input is not None and not args.languages:
            parser.error('When --input is specified, each column\'s LANG has to be specified as well')

        if args.tee and args.parallel > 1:
            parser.error('Using --tee is not supported when using --parallel')

        if args.time and not args.trace:
            parser.error('You need to use --trace to see the output of --time')

        # load all filter definitions (we need to, to get their name)
        filters = {
            definition.name: definition
            for definition in list_filters(args.filters)
        }

        # set_global_filters() provides the filters to the validators in FilterPipeline
        set_global_filters(filters.values())
        pipeline_config = parse_obj_as(FilterPipeline, json.load(args.pipeline))

        # Order of columns. Matches datasets.py:list_datasets(path)
        languages: List[str] = args.languages if args.input else [filename.rsplit('.', 2)[1] for filename in pipeline_config.files]

        # Directory plus basename to write debug (`--tee`) files to
        basename: str = 'stdin' if args.input else os.path.commonprefix(pipeline_config.files).rstrip('.')

        pipeline = Pipeline(filters, languages, pipeline_config)

        # Input for next child
        stdin: IO[bytes]

        # Output of this program
        stdout:IO[bytes] = args.output

        # If we're just dumping the pipeline, do so to the specified output
        if args.dump:
            pipeline.dump(TextIOWrapper(stdout))
            sys.exit(0)

        # Queue filled by the babysitters with the stderr of the children, consumed
        # by `print_lines()` to prevent racing on stderr.
        print_queue = SimpleQueue() # type: SimpleQueue[Optional[bytes]]

        # First start the print thread so that we get immediate feedback from the
        # children even if all of them haven't started yet.
        print_thread = Thread(target=print_lines, args=[print_queue, sys.stderr.buffer])
        print_thread.start()

        # Start child processes, each reading the output from the previous sibling
        try:
            with ProcessPool(print_queue) as pool:
                # If we're not reading from stdin, read from files and paste them together
                if args.input:
                    stdin = args.input
                else:
                    # Open `gzunip` for each language file
                    gunzips = [
                        pool.start(f'gunzip {filename}',
                            ['gzip', '-cd', filename],
                            stdout=PIPE,
                            stderr=PIPE,
                            cwd=args.basedir)
                        for filename in pipeline_config.files
                    ]

                    fds = [none_throws(gunzip.stdout).fileno() for gunzip in gunzips]

                    # .. and a `paste` to combine them into columns
                    paste = pool.start('paste',
                        ['paste'] + [f'/dev/fd/{fd}' for fd in fds],
                        stdout=PIPE,
                        stderr=PIPE,
                        pass_fds=fds)

                    # Now that `paste` has inherited all the children, close our connection to them
                    for gunzip in gunzips:
                        none_throws(gunzip.stdout).close()

                    stdin = none_throws(paste.stdout)

                # If we only want the first N lines processed, use `head` to chop those off.
                if args.first > 0:
                    head = pool.start('head',
                        ['head', '-n', str(args.first)],
                        stdin=stdin,
                        stdout=PIPE,
                        stderr=PIPE)

                    stdin.close() # now taken over by `head`.
                    stdin = none_throws(head.stdout)

                if args.parallel > 1:
                    run_parallel(pipeline, stdin, stdout, print_queue=print_queue, parallel=args.parallel, batch_size=args.batch_size, time=args.time)
                else:
                    pipeline.run(pool, stdin, stdout, tee=args.tee, basename=basename, time=args.time)
        except:
            # If we didn't cleanly exit all processes, we err as well
            traceback.print_exc(file=sys.stderr)
            sys.exit(1)
        finally:
            # Tell print thread to stop (there are no more babysitters now to produce printable stuff)
            print_queue.put(None)
            print_thread.join()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""Legacy fix_un.py from Barry. Need to fix it up a bit."""
import re
import sys


re_final_comma = re.compile(r"\.$")


for line in sys.stdin:
    line = line.strip()
    if line[-1] == '，':
        line = line[:-1] + "\u3002"
    line = line.replace(",", "\uFF0C")
    line = re_final_comma.sub("\u3002", line)
    print(line)



#!/usr/bin/env python3
import sys
import argparse
from typing import TextIO


def parse_user_args():
    """Parse the arguments necessary for this filter"""
    parser = argparse.ArgumentParser(description="Filters a parallel dataset based on max word length")
    parser.add_argument("--max-word-length", default=150, type=int)
    return parser.parse_args()


def clean_parallel(max_word_length: float, fin: TextIO, fout: TextIO) -> None:
    """Cleans the parallel or mono dataset based on line lengths"""
    for line in fin:
        fields = line.rstrip('\r\n').split('\t')

        if any(len(token) > max_word_length
            for field in fields
            for token in field.split(' ')):
            continue

        fout.write(line)


if __name__ == '__main__':
    args = parse_user_args()
    clean_parallel(args.max_word_length, sys.stdin, sys.stdout)



#!/usr/bin/env python3
import sys
import re
import argparse
from enum import Enum
from typing import Set, List, Tuple


class MatchType(Enum):
	EXACT = 'exact'
	COUNT = 'count'


Pattern = Tuple[str,str, MatchType]

# Footnote pattern needs to have the exact same matches on both sides. They're
# just things like `[3]` at the end of a word.
FOOTNOTE_PATTERN: Pattern = r'\[[0-9]+\]', r'', MatchType.EXACT

# URL match needs to be an exact match on both sides. If not, we teach the MT
# system to translate urls for us and that's not good.
URL_PATTERN: Pattern = r'https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&//=]*)', r'', MatchType.EXACT

# If wiki links, e.g. `[[pagename|Link Label]]` or ``[[pagename]]` don't appear
# in equal amount on both sides, replace them with just their label. Not
# comparing the links themselves since they may link to translated page names.
WIKILINKS_PATTERN: Pattern = r'\[\[(?:.+?\|)?(.+?)\]\]', r'\1', MatchType.COUNT

# If header does not appear on both sides, remove it entirely. We only compare
# counts because the headers themselves are probably translated.
HEADINGS_PATTERN: Pattern = r'(==+)(.+?)\1', r'', MatchType.COUNT

CODE_PATTERN = r'\.mw-parser-output' # Very specific for OPUS-wikimedia


def find_matches(pattern:Pattern, text:str) -> Set[str]:
	return set(match[0] for match in re.finditer(pattern[0], text))


def filter_matches(pattern:Pattern, text:str) -> str:
	return re.sub(pattern[0], pattern[1], text)


def is_mismatch(pattern:Pattern, fields: List[str]) -> bool:
	matches = [find_matches(pattern, field) for field in fields[:2]]
	if pattern[2] == MatchType.EXACT:
		return len(matches[0] & matches[1]) < len(matches[0] ^ matches[1])
	elif pattern[2] == MatchType.COUNT:
		return len(matches[0]) != len(matches[1])
	else:
		raise NotImplementedError()


def is_code(field: str) -> bool:
	return re.search(CODE_PATTERN, field) is not None


if __name__ == "__main__":
	parser = argparse.ArgumentParser(description="Remove common wiki patterns from sentence pair if they don't match on both sides")
	parser.add_argument("--always", action="store_true", help="Always remove patterns")
	parser.add_argument("--footnotes", action="store_true", help="Remove footnotes, e.g. [1], [2]")
	parser.add_argument("--urls", action="store_true", help="Remove url`s")
	parser.add_argument("--wikilinks", action="store_true", help="Remove [[wikilinks]]")
	parser.add_argument("--code", action="store_true", help="Remove lines that contain code")
	parser.add_argument("--headings", action="store_true", help="Remove ==headings==")
	parser.add_argument("--remove-empty-lines", action="store_true", help="Remove sentence pairs when one side is empty after filtering")
	args = parser.parse_args()

	patterns: List[Pattern] = []

	if args.footnotes:
		patterns.append(FOOTNOTE_PATTERN)

	if args.urls:
		patterns.append(URL_PATTERN)

	if args.wikilinks:
		patterns.append(WIKILINKS_PATTERN)

	if args.headings:
		patterns.append(HEADINGS_PATTERN)

	for n, line in enumerate(sys.stdin, start=1):
		fields = line.rstrip("\r\n").split("\t")

		if args.code and any(is_code(field) for field in fields):
			continue

		for pattern in patterns:
			if args.always or is_mismatch(pattern, fields[:2]):
				fields[:2] = [filter_matches(pattern, field) for field in fields[:2]]

		# Make sure we didn't add padding after all that replacing
		fields = [field.strip() for field in fields]

		if args.remove_empty_lines and all(len(field) > 0 for field in fields):
			print("\t".join(fields))



#!/usr/bin/env python3
import sys

my_punct = {'!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '.', '/', ':', ';', '<', '=', '>', '?', '@', '[', '\\', ']', '^', '_', '`', '{', '|', '}', '~', '»', '«', '“', '”'}

for line in sys.stdin:
    src, trg = line.rstrip("\r\n").split("\t")
    if len(src) == 0 or len(trg) == 0:
        print(src + '\t' + trg)
        continue

    # Sometimes we have a space between the final letter and the punctuation, which is wrong except if using french quotes
    if len(src) >= 2 and src[-1] in my_punct and src[-2] == " " and src[-1] != '»' and src[-1] != '«':
        src = src[:-2] + src[-1]
    if len(trg) >= 2 and trg[-1] in my_punct and trg[-2] == " " and trg[-1] != '»' and trg[-1] != '«':
        trg = trg[:-2] + trg[-1]
    # Sometimes two punctuation marks are swapped...
    if len(src) >=2 and len(trg) >= 2 and src[-2] == trg[-1] and src[-1] == trg[-2]:
       trg = trg[:-2] + src[-2] + src[-1]
    # Sometimes they are swapped with space around eg SPACE». -> .SPACE»
    if len(src) >=3 and src[-1] in my_punct and src[-2] == '»' and src[-3] == ' ':
       src = src[:-3] + src[-1] + ' ' + src[-2]
    if len(trg) >=3 and trg[-1] in my_punct and trg[-2] == '»' and trg[-3] == ' ':
       trg = trg[:-3] + trg[-1] + ' ' + trg[-2]


    # check for the french quotes special case
    if (src[-1] == '»' or src[-1] == '«') and trg[-1] not in my_punct:
        trg = trg + '"'
    elif (trg[-1] == '»' or trg[-1] == '«') and src[-1] not in my_punct:
        src = src + '"'
    elif src[-1] in my_punct and trg[-1] not in my_punct:
        trg = trg + src[-1]
    elif trg[-1] in my_punct and src[-1] not in my_punct:
        src = src + trg[-1]
    # Final case. Fix mismatched punctuation on the src and trg. EXCEPT in cases like french quotes. And in cases where we have emdash at the front, as it means spech
    elif trg[-1] in my_punct and src[-1] in my_punct and src[-1] != trg[-1] and src[-1] != '»' \
and src[-1] != '«' and trg[-1] != '»' and trg[-1] != '«' and src[0] != '–' and trg[0] != '–' and src[0] != '—' and trg[0] != '—':
        trg = trg[:-1] + src[-1]
    print(src + '\t' + trg)



#!/usr/bin/env python3
import sys

QUOTECHR = ord('"')

for line in sys.stdin.buffer:
	fields = line.rstrip(b"\r\n").split(b"\t")
	for i, field in enumerate(fields):
		if len(field) > 0 and field[0] == QUOTECHR and field[-1] == QUOTECHR:
			fields[i] = field[1:-1].replace(b'""', b'"')
	sys.stdout.buffer.write(b"\t".join(fields))
	sys.stdout.buffer.write(b"\n")



#!/usr/bin/env python3
import os
import sys
import argparse
from typing import List

# The filename has to have a different name than 'fasttext', otherwise we can't import the module correctly
import fasttext

from more_itertools import chunked
import requests


def download_model(model_type: str):
    """
    Downloads the fasttext model for language identification to local directory. Either the large one or the small one.
    """
    if model_type == "small":
        url = "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.ftz"
    elif model_type == "large":
        url = "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"
    else:
        raise TypeError("Fasttext model type has to be either 'small' or 'large'.")
    file_name = model_type + ".bin"
    # Do not download twice
    if os.path.exists(file_name):
        return
    response = requests.get(url)
    with open(file_name, "wb") as handle:
        handle.write(response.content)


def verify_lang(model: fasttext.FastText._FastText, texts: List[str], desired_lang: str, debug: bool) -> List[bool]:
    # Langs is a list of list - for each row we get a list of identified languages, sorted by their probability.
    # Future work - using `model.predict(texts, k=10)` get the 10 most probable languages
    #   and do some clever filtering based on the distribution.
    langs, probs = model.predict(texts)

    if debug:
        sys.stderr.write(f"LANGUAGES\t{[lang[0] for lang in langs]}\n")
    return [row_langs[0] == desired_lang for row_langs in langs]


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--source-lang", type=str, help="Code of the desired source language.")
    parser.add_argument("--target-lang", type=str, help="Code of the desired target language.")
    parser.add_argument("--batch-size", type=int, default=16, help="Size of the batch to send the data to fasttext.")
    parser.add_argument("--debug", action="store_true")
    parser.add_argument(
        "--model-type",
        type=str,
        help="Either 'small' or 'large'. See https://fasttext.cc/docs/en/language-identification.html for more info.",
    )
    args = parser.parse_args()
    # Fastext way to encode language codes
    source_lang = "__label__" + args.source_lang
    target_lang = "__label__" + args.target_lang
    download_model(args.model_type)
    # Disable fasttext to notify us about loading the model
    fasttext.FastText.eprint = lambda x: None
    model = fasttext.load_model(f"{args.model_type}.bin")
    for batch in chunked(sys.stdin, args.batch_size):
        # Remove newlines
        batch = [row.rstrip("\r\n") for row in batch]
        sources, targets = zip(*[row.split("\t", 1) for row in batch])
        sources_ok = verify_lang(model, list(sources), source_lang, args.debug)
        targets_ok = verify_lang(model, list(targets), target_lang, args.debug)
        for row, source_ok, target_ok in zip(batch, sources_ok, targets_ok):
            if source_ok and target_ok:
                sys.stdout.write(row + "\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from sys import stdin, stdout, stderr
import argparse


def parse_user_args():
    """Parse the arguments necessary for this filter"""
    parser = argparse.ArgumentParser(description="Filters a parallel or mono dataset based on line lengths")
    parser.add_argument("--max-length", default=150, type=float)
    parser.add_argument("--min-length", default=1, type=float)
    parser.add_argument("--debug", action='store_true')
    return parser.parse_args()


def clean_parallel(max_length: float, min_length: float, debug: bool=True) -> None:
    """Cleans the parallel or mono dataset based on line lengths"""
    for line in stdin:
        fields = line.strip('\r\n').split('\t')
        if len(fields) == 1:
            src = fields[-1].strip()
            trg = None
        else: # Assumes that the multiline filter already run
            src = fields[-2].strip()
            trg = fields[-1].strip()

        srctok = src.split()
        srcpass: bool = (len(srctok) <= max_length and len(srctok) >= min_length)

        trgpass: bool

        # Check lengths
        if trg is None:
            trgpass = True
        else:
            trgtok = trg.split()
            trgpass = (len(trgtok) <= max_length and len(trgtok) >= min_length)

        # write
        if srcpass and trgpass:
            stdout.write(line)
        elif debug:
            stderr.write(f'LENGTH\t{src}\t{trg}\n')


if __name__ == '__main__':
    args = parse_user_args()
    clean_parallel(args.max_length, args.min_length, args.debug)



#!/usr/bin/env python3
import sys
import argparse
from typing import TextIO, List
from sentence_splitter import SentenceSplitter


def split_sentences_in_bitext(fin: TextIO, fout: TextIO, languages: List[str], keep_unbalanced: bool = False):
	splitters = [SentenceSplitter(language=lang) for lang in languages]
	
	for line in fin:
		cols = line.rstrip('\r\n').split('\t')

		assert len(cols) == len(splitters)

		splitted = [splitter.split(col) for splitter, col in zip(splitters, cols)]

		if any(len(col) != len(splitted[0]) for col in splitted[1:]):
			if keep_unbalanced:
				# Revert back to the input line
				splitted = [[col] for col in cols]
			else:
				# Skip line
				continue

		for cols in zip(*splitted):
			fout.write('\t'.join(cols) + '\n')


if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('--keep-unbalanced', action='store_true')
	parser.add_argument('languages', type=str, nargs='+')

	args = parser.parse_args()

	split_sentences_in_bitext(sys.stdin, sys.stdout, args.languages, args.keep_unbalanced)



#!/usr/bin/env python3
"""Various mtdata dataset downloading utilities"""
import os
import sys
import asyncio
import json
import gzip
from glob import iglob
from itertools import chain
from typing import Iterable, Dict, List, Optional, Set, Union, Tuple, cast, Any
from enum import Enum
from queue import SimpleQueue
from subprocess import Popen, PIPE
from threading import Thread
from collections import defaultdict
from urllib.request import Request, urlopen
from urllib.parse import urlencode
from pprint import pprint
from operator import itemgetter
from warnings import warn
from tempfile import TemporaryDirectory, TemporaryFile, NamedTemporaryFile
from shutil import copyfileobj
from multiprocessing import Process
from zipfile import ZipFile
from concurrent.futures import ProcessPoolExecutor

from pydantic import BaseModel
from fastapi import FastAPI, HTTPException

from opuscleaner.config import DATA_PATH, DOWNLOAD_PATH


class EntryRef(BaseModel):
    id: int


class Entry(EntryRef):
    corpus: str
    version: str
    langs: Tuple[str,str]
    pairs: Optional[int] # Number of sentence pairs
    size: Optional[int] # Size on disk in bytes (rounded to lowest 1024)

    @property
    def basename(self) -> str:
        return f'{self.corpus}-{self.version}.{"-".join(self.langs)}'


class LocalEntry(Entry):
    paths: Set[str]


class RemoteEntry(Entry):
    url: str


class DownloadState(Enum):
    PENDING = 'pending'
    CANCELLED = 'cancelled'
    DOWNLOADING = 'downloading'
    DOWNLOADED = 'downloaded'
    FAILED = 'failed'


def get_dataset(entry:RemoteEntry, path:str) -> None:
    if entry.url.endswith('.zip'):
        get_bilingual_dataset(entry, path)
    elif entry.url.endswith('.txt.gz'):
        get_monolingual_dataset(entry, path)
    else:
        raise RuntimeError(f'Unknown dataset file type: {entry.url}')


def get_monolingual_dataset(entry:RemoteEntry, path:str) -> None:
    lang = next(lang for lang in entry.langs if lang != '')
    assert entry.url.endswith(f'{lang}.txt.gz')
    
    # Make sure our path exists
    os.makedirs(path, exist_ok=True)

    with TemporaryDirectory(dir=path) as temp_dir:
        temp_path = os.path.join(temp_dir, os.path.basename(entry.url))
        dest_path = os.path.join(path, f'{entry.basename}.{lang}.gz')

        # Download dataset to temporary file
        with urlopen(entry.url) as fh, open(temp_path, 'wb') as fout:
            copyfileobj(fh, fout)

        # move to permanent position
        os.rename(temp_path, dest_path)


def _extract(path:str, name:str, dest:str) -> str:
    with ZipFile(path) as archive, archive.open(name) as fin, gzip.open(dest, 'wb') as fout:
        copyfileobj(fin, fout, length=2**24) # 16MB blocks
    return dest


def get_bilingual_dataset(entry:RemoteEntry, path:str) -> None:
    # List of extensions of the expected files, e.g. `.en-mt.mt` and `.en-mt.en`.
    suffixes = [f'.{"-".join(entry.langs)}.{lang}' for lang in entry.langs]

    # Make sure our path exists
    os.makedirs(path, exist_ok=True)

    with NamedTemporaryFile() as temp_archive:
        # Download zip file to temporary file
        with urlopen(entry.url) as fh:
            copyfileobj(fh, temp_archive)

        # Then selectively extract that zipfile to a temporary directory
        with TemporaryDirectory(dir=path) as temp_extracted:
            files = []

            with ProcessPoolExecutor(max_workers=8) as pool:
                futures = []

                with ZipFile(temp_archive) as archive:
                    for info in archive.filelist:
                        if info.is_dir() or not any(info.filename.endswith(suffix) for suffix in suffixes):
                            continue

                        # `info.filename` is something like "beepboop.en-nl.en", `lang` will be "en".
                        _, lang = info.filename.rsplit('.', maxsplit=1)

                        filename = f'{entry.basename}.{lang}.gz'
                        temp_dest = os.path.join(temp_extracted, filename)
                        data_dest = os.path.join(path, filename)

                        # Extract the file from the zip archive into the temporary directory, and
                        # compress it while we're at it.
                        future = pool.submit(_extract, temp_archive.name, info.filename, temp_dest)

                        futures.append((future, data_dest))

                # Keep a list of extracted files, and where they eventually need to go to
                for future, data_dest in futures:
                    files.append((future.result(), data_dest))

            # Once we know all files extracted as expected, move them to their permanent place.
            for temp_path, dest_path in files:
                os.rename(temp_path, dest_path)


class EntryDownload:
    entry: RemoteEntry
    _child: Optional[Process]

    def __init__(self, entry:RemoteEntry):
        self.entry = entry
        self._child = None
    
    def start(self) -> None:
        self._child = Process(target=get_dataset, args=(self.entry, DOWNLOAD_PATH))
        self._child.start()

    def run(self) -> None:
        self.start()
        assert self._child is not None
        self._child.join()

    def cancel(self) -> None:
        if self._child and self._child.exitcode is None:
            self._child.kill()

    @property
    def state(self) -> DownloadState:
        if not self._child:
            return DownloadState.PENDING
        elif self._child.exitcode is None:
            return DownloadState.DOWNLOADING
        elif self._child.exitcode == 0:
            return DownloadState.DOWNLOADED
        elif self._child.exitcode > 0:
            return DownloadState.FAILED
        else:
            return DownloadState.CANCELLED


DownloadQueue = SimpleQueue#[Optional[EntryDownload]]


class Downloader:
    def __init__(self, workers:int):
        self.queue: DownloadQueue = SimpleQueue()
        self.threads: List[Thread] = []

        for _ in range(workers):
            thread = Thread(target=self.__class__.worker_thread, args=[self.queue], daemon=True)
            thread.start()
            self.threads.append(thread)

    def download(self, entry:RemoteEntry) -> EntryDownload:
        download = EntryDownload(entry=entry)
        self.queue.put(download)
        return download

    @staticmethod
    def worker_thread(queue:DownloadQueue) -> None:
        while True:
            entry = queue.get()
            if not entry:
                break
            entry.run()


class EntryDownloadView(BaseModel):
    entry: Entry
    state: DownloadState


class OpusAPI:
    endpoint: str

    _datasets: Dict[int,Entry] = {}

    def __init__(self, endpoint:str):
        self.endpoint = endpoint
        self._datasets = {}

    def languages(self, lang1: Optional[str] = None) -> List[str]:
        query = {'languages': 'True'}

        if lang1 is not None:
            query['source'] = lang1

        with urlopen(f'{self.endpoint}?{urlencode(query)}') as fh:
            return [str(lang) for lang in json.load(fh).get('languages', [])]

    def get_dataset(self, id:int) -> Entry:
        return self._datasets[id]

    def find_datasets(self, lang1:str, lang2:Optional[str]=None) -> List[Entry]:
        if lang2 is None:
            query = {
                'source': lang1,
                'preprocessing': 'mono'
            }
        else:
            query = {
                'source': lang1,
                'target': lang2,
                'preprocessing': 'moses'
            }

        with urlopen(f'{self.endpoint}?{urlencode(query)}') as fh:
            datasets = [cast_entry(entry) for entry in json.load(fh).get('corpora', [])]

        # FIXME dirty hack to keep a local copy to be able to do id based lookup
        # Related: https://github.com/Helsinki-NLP/OPUS-API/issues/3
        for dataset in datasets:
            self._datasets[dataset.id] = dataset

        return datasets


app = FastAPI()

api = OpusAPI('https://opus.nlpl.eu/opusapi/')

downloads: Dict[int,EntryDownload] = {}

downloader = Downloader(2)

datasets_by_id: Dict[int, Entry] = {}

def cast_entry(data:Dict[str,Any]) -> Entry:
    entry = Entry(
        id=int(data['id']),
        corpus=str(data['corpus']),
        version=str(data['version']),
        pairs=int(data['alignment_pairs']) if data.get('alignment_pairs') != '' else None,
        size=int(data['size']) * 1024, # FIXME file size but do we care?
        langs=(data['source'], data['target']), # FIXME these are messy OPUS-API lang codes :(
    )

    paths = set(
        filename
        for data_root in [os.path.dirname(DATA_PATH), DOWNLOAD_PATH]
        for lang in entry.langs
        for filename in iglob(os.path.join(data_root, f'{entry.basename}.{lang}.gz'), recursive=True)
    )

    # Print search paths
    # print('\n'.join(
    #     filename
    #     for data_root in [os.path.dirname(DATA_PATH), DOWNLOAD_PATH]
    #     for lang in cast(Tuple[str,str], entry.langs)
    #     for filename in [os.path.join(data_root, f'{entry.basename}.{lang}.gz')]
    # ))

    if paths:
        return LocalEntry(
            **entry.__dict__,
            paths=paths)
    else:
        return RemoteEntry(
            **entry.__dict__,
            url=str(data['url']))


@app.get("/languages/")
@app.get("/languages/{lang1}")
def list_languages(lang1:Optional[str] = None) -> List[str]:
    return sorted(api.languages(lang1))


@app.get("/by-language/{langs}")
def list_datasets(langs:str) -> Iterable[Entry]:
    return api.find_datasets(*langs.split('-'))


@app.get('/downloads/')
def list_downloads() -> Iterable[EntryDownloadView]:
    return (
        EntryDownloadView(
            entry = download.entry,
            state = download.state
        )
        for download in downloads.values()
    )


@app.post('/downloads/')
def batch_add_downloads(datasets: List[EntryRef]) -> Iterable[EntryDownloadView]:
    """Batch download requests!"""
    needles = set(dataset.id
        for dataset in datasets
        if dataset.id not in downloads
        or downloads[dataset.id].state in {DownloadState.CANCELLED, DownloadState.FAILED})

    entries = [
        api.get_dataset(id) for id in needles
    ]

    for entry in entries:
        assert isinstance(entry, RemoteEntry)
        downloads[entry.id] = downloader.download(entry)

    return list_downloads()


@app.delete('/downloads/{dataset_id}')
def cancel_download(dataset_id:int) -> EntryDownloadView:
    """Cancel a download. Removes it from the queue, does not kill the process
    if download is already happening.
    """
    if dataset_id not in downloads:
        raise HTTPException(status_code=404, detail='Download not found')

    download = downloads[dataset_id]
    download.cancel()

    return EntryDownloadView(
        entry = download.entry,
        state = download.state
    )



#!/usr/bin/env python3
import argparse
import importlib
import itertools
import logging
import os
import sys
import warnings

# Prevent searching for modules in the filters/ directory (like langid)
del sys.path[0]

import opusfilter
import yaml

parser = argparse.ArgumentParser()
parser.add_argument('--quiet', '-q', action='store_true')
parser.add_argument('filter', type=str)
parser.add_argument('config', type=str)

args = parser.parse_args()

if args.quiet:
	# Filter out warnings from opusfilter about missing env variables for like
	# word alignment.
	logging.getLogger().setLevel(logging.ERROR)

	# Filter warnings (especially MarkupResemblesLocatorWarning) from
	# BeautifulSoup (the HtmlTagFilter)
	warnings.filterwarnings('ignore', module='bs4')

module_path, class_name = args.filter.rsplit('.', maxsplit=1)

config = yaml.safe_load(args.config)

if not isinstance(config, dict):
	if config:
		raise ValueError('config has to be a mapping')

	config = dict()

mod = importlib.import_module(module_path)
filter_cls = getattr(mod, class_name)
filter_obj = filter_cls(**config)

if isinstance(filter_obj, opusfilter.FilterABC):
	def apply_filter(lines):
		# Duplicate the iterator into two, one goes into the scorer, one for output
		# because scorer could be eating them in chunks.
		lines1, lines2 = itertools.tee(lines)
		pairs = (line[0:2] for line in lines1)
		for line, score in zip(lines2, filter_obj.score(pairs)):
			if filter_obj.accept(score):
				yield line
elif isinstance(filter_obj, opusfilter.PreprocessorABC):
	def apply_filter(pairs):
		return filter_obj.process(pairs)
else:
	raise ValueError('filter class does not implement FilterABC or PreprocessorABC')

lines = (line.rstrip('\r\n').split('\t') for line in sys.stdin)

for line in apply_filter(lines):
	print("\t".join(line))



#!/usr/bin/env python3
"""
Removes patterns that occur usually only in the source or in the target. There are two way to set up a pattern:
 - a two columns type - pattern tab replacement, which is applied to both source and target
 - a three columns tpye - pattern_on_both_cols tab pattern tab replacement
    The `pattern_on_both_cols` is matched against tab separated source and target.
    If it matches then the `pattern` is replaced in both columns.

You can check the three column filter from the examples by running
```
mtdata get -l ces-eng -tr ELRC-czech_supreme_audit_office_2018_reports-1-ces-eng  --compress -o data
```
and using the `remove_frequent_patterns` filter.
Then e.g. the line containing `Increasing support for informal carers,` will not start with `• `,
because the source doesn't start with a bullet either.
"""

from dataclasses import dataclass
import re
import sys
import argparse
from typing import List, Optional


@dataclass
class Pattern:
    group_match: re.Pattern
    replacement: str
    pattern_on_both_cols: Optional[re.Pattern] = None


def load_patterns(file_path: str) -> List[Pattern]:
    with open(file_path) as f:
        # Skip comment lines
        lines = [line.rstrip("\r\n") for line in f.readlines() if line[0] != "#"]
        patterns = []
        for line in lines:
            parts = line.split("\t")
            if len(parts) == 2:
                patterns.append(Pattern(group_match=re.compile(parts[0]), replacement=parts[1]))
            elif len(parts) == 3:
                patterns.append(Pattern(pattern_on_both_cols=re.compile(parts[0]), group_match=re.compile(parts[1]), replacement=parts[2]))
            else:
                raise ValueError(f"Patterns have to have 2 or 3 columns, but got {len(parts)}")
        return patterns


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--pattern-file", type=str, help="Path to the file with patterns.")
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()
    patterns = load_patterns(args.pattern_file)

    for line in sys.stdin:
        line = line.rstrip("\r\n")
        source, target = line.split("\t", 1)
        for pattern in patterns:
            # Either pattern_on_both_cols is not set or it matches the whole line
            if pattern.pattern_on_both_cols is None or pattern.pattern_on_both_cols.match(line):
                source = pattern.group_match.sub(pattern.replacement, source)
                target = pattern.group_match.sub(pattern.replacement, target)
        sys.stdout.write(f"{source}\t{target}\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
from sys import stdin, stdout, stderr
from typing import BinaryIO
import argparse

import pycld2

# Similar languages, taken from 
# https://github.com/mbanon/fastspell/blob/main/fastspell/config/similar.yaml
SIMILAR = {
    "ca": {"es", "ca"},
    "bg": {"mk", "bg"},
    "bs": {"bs", "hr", "me", "sr"},
    "cs": {"sk", "cs"},
    "da": {"da", "nb"},
    "es": {"es", "gl", "ca"},
    "gl": {"es", "pt", "gl"},
    "hr": {"bs", "hr", "me", "sr"},
    "me": {"bs", "hr", "me", "sr"},
    "mk": {"bg", "mk"},
    "nb": {"nn", "da", "nb"},
    "nl": {"nl", "af"}, # Maybe also Frisian (fy) and French (fr) because of
                        # short sentences are often misidentified as one of
                        # those (and honestly cld2 has probably been trained
                        # with a lot of Dutch in their Frisian corpora.)
    "nn": {"nb", "da", "nn"},
    "sk": {"cs", "sk"},
    "sr": {"bs", "hr", "me", "sr"},
}

LANG_UNKNOWN = "un"

def parse_user_args():
    """Parse the arguments necessary for this filter"""
    parser = argparse.ArgumentParser(description="Langid")
    parser.add_argument("languages", default=None, type=str, nargs="+")
    parser.add_argument("--allow-similar", action="store_true")
    parser.add_argument("--allow-unknown", action="store_true")
    parser.add_argument("--debug", action="store_true")
    return parser.parse_args()


def detect_language_parallel(args: argparse.Namespace, fin: BinaryIO, fout: BinaryIO):
    for n, line in enumerate(fin):
        fields = line.rstrip(b"\r\n").split(b"\t")

        for field, lang in zip(fields, args.languages):
            try:
                isReliable, _, [(_, detected_lang, _, _), *_] = pycld2.detect(field)

                # Accept field if reliable and detected language
                if isReliable and detected_lang == lang:
                    continue

                # Accept field if reliable and similar language detected
                if isReliable and args.allow_similar and detected_lang in SIMILAR[lang]:
                    continue

                # Accept field if not reliable but we allow failed detections
                if not isReliable and args.allow_unknown:
                    continue

                if args.debug:
                    print(f"Line {n} rejected. Detected '{detected_lang}', expected '{lang}': {field.decode()}", file=stderr)
                
                # Break because no need to look at the other columns. Also will
                # stop the else clause from being executed!
                break
            except Exception as err:
                if not args.allow_unknown:
                    print(f"Line {n} rejected because error: {err}", file=stderr)
                    break
        else:
            # All fields were valid, so write out the line.
            fout.write(line)
            continue


if __name__ == "__main__":
    args = parse_user_args()
    detect_language_parallel(args, stdin.buffer, stdout.buffer)



#!/usr/bin/env python3
"""Common filtering code to be used by various submodules"""

CHARS = {
    'ar': r'[\u0600-\u06FF]', # This is not entirely right, as it also includes farsi symbols and whatnot
    'bg': r'[АаБбВвГгДддЕеЖжЗзИиЙйКкkasЛлМмНнОоПпРрСсТтУуФфХхЦцЧчШшЩщЪъЬьЮюЯя]',
    'bn': r'[\u0980-\u09FF]', # bangla
    'ca': r'[a-zÀàÈèÉéÍíÒòÓóÚúÇç]',
    'cs': r'[a-zÁáČčĎďÉéěÍíŇňÓóŘřŠšŤťÚúůÝýŽž]',
    'da': r'[a-zÆæØøÅå]',
    'de': r'[a-zÄäÖöÜüß]',
    'en': r'[a-z]',
    'el': r'[a-zΑαΒβΓγΔδΕεΖζΗηΘθΙιΚκΛλΜμΝνΞξΟοΠπΡρΣσςΤτΥυΦφΧχΨψΩω]',
    'es': r'[a-zÁáÉéÍíÓóÚúñÑ]',
    'et': r'[a-zÕõÄäÖöÜü]',
    'eu': r'[a-zñÑ]',
    'fi': r'[a-zÅåÄäÖö]',
    'fr': r'[a-zÂâÁáÀàâÇçÉéÈèÊêÓóÒòÔôŒœÜüÛûŸÿ]',
    'ga': r'[abcdefghilmnoprstuáéíóúÁÉÍÓÚ]',
    'gl': r'[a-zÁáÉéÍíÓóÚúÑñ]',
    'hi': r'[\u0900-\u097F]', # devanagari
    'hr': r'[abcčČćĆdđĐefghijklmnoprsšŠtuvzžŽ]',
    'hu': r'[a-zÁáÉéÍíÓóÖöŐőŰű]',
    'hy': r'[\u0530-\u058F]',
    'is': r'[abdefghijklmnoprstuvxyÁáðÐÉéÍíÓóÚúÝýÞþÆæÖö]',
    'it': r'[a-zàÀèÈéÉìÌíÍîÎòÒóÓùÙúÚ]',
    'ko': r'[\uac00-\ud7af]|[\u1100-\u11ff]|[\u3130-\u318f]|[\ua960-\ua97f]|[\ud7b0-\ud7ff]',
    'lt': r'[aąbcČčdeĘęĖėfghiĮįyjklmnoprsŠštuŲųŪūvzŽž]',
    'lv': r'[aĀābcČčdeĒēfgĢģhiĪījkĶķlĻļmnŅņoprsŠštuŪūvzŽž]',
    'mt': r'[abĊċdefĠġghĦħiiejklmnopqrstuvwxŻżz]',
    'nb': r'[a-zÂâÁáÀàâÉéÈèÊêÓóÒòÔôÜüÆæØøÅå]',
    'nl': r'[a-zÂâÁáÀàâÉéÈèÊêÓóÒòÔôÚú]',
    'no': r'[a-zÂâÁáÀàâÉéÈèÊêÓóÒòÔôÜüÆæØøÅå]',
    'nn': r'[a-zÂâÁáÀàâÉéÈèÊêÓóÒòÔôÜüÆæØøÅå]',
    'pl': r'[a-zĄąĆćĘęŁłŃńÓóŚśŹźŻż]',
    'pt': r'[a-zÂâÁáÀàÃãÇçÉéÈèÊêÍíÌìÓóÒòÔôÕõÚúÙù]',
    'ro': r'[a-zĂăÂâÎîȘșȚț]',
    'ru': r'[а-я]',
    'sk': r'[a-záäÁÄčČďĎžéÉíÍĺĹľĽňŇóÓôÔŕŔšŠťŤúÚýÝžŽ]',
    'sl': r'[abcčČdđĐefghijklmnoprsšŠtuvzžŽ]',
    'sv': r'[a-zÅåÄäÖö]',
    'uk': r'[А-ЩЬЮЯҐЄІЇа-щьюяґєії\'`’ʼ]',
    'zh': r'[\u3400-\u4dbf\u4e00-\u9fff\uf900-\ufaff]',
}



#!/usr/bin/env python3
"""Compute score and optionally threshold and cache per line of input.

It passes every line of input onto the scorer program (unless --cache is
specified and the line is already in the cache) that generates a score. If
--threshold is specified (optionally with an operator specified, default is 
greater or equal, c.f. --ge) then the line is printed if the threshold is met.
If no threshold is specified, the score is added as the first column to the
output.

Note on --cache: make sure that if you change the arguments to the scorer
program, you also change the path to the cache. Otherwise you'll get scores
from the scorer that was run with different arguments.
"""
import sys
import os
import signal
import argparse
import dbm
import operator
from xxhash import xxh32
from traceback import print_exc
from subprocess import Popen, PIPE
from threading import Thread
from queue import SimpleQueue
from typing import Optional, TypeVar, Dict
from functools import wraps
import struct


class Entry:
	"""Cache entry. Only an object with single property so we can easily update
	it by reference, really.
	"""
	__slots__ = ['score']
	score: Optional[float]

	def __init__(self, score: Optional[float] = None):
		self.score = score


class Cache:
	"""Just a subset of dict[] really!"""
	def __init__(self):
		self.entries = {}

	def __contains__(self, key: bytes) -> bool:
		return key in self.entries
	
	def __getitem__(self, key: bytes) -> Entry:
		return self.entries[key]

	def __setitem__(self, key: bytes, value: Entry):
		self.entries[key] = value

	def __enter__(self):
		return self

	def __exit__(self, *args):
		return


class PersistentEntry:
	"""Mimics Entry(), but with a setter that updates the persistent cache."""
	__slots__ = ['cache', 'key', '_score']
	cache: "PersistentCache"
	key: bytes

	def __init__(self, cache, key: bytes, score: Optional[float] = None):
		self.cache = cache
		self.key = key
		self._score = score

	@property
	def score(self) -> Optional[float]:
		return self._score

	@score.setter
	def score(self, score: float):
		self._score = score
		self.cache._write(self.key, score)


class PersistentCache(Cache):
	"""Similar to Cache, but will also look at the database file that's on disk
	when queried for known entries.
	"""
	__slots__ = ['entries', 'db', '_backing']

	def __init__(self, path: str):
		self.db = dbm.open(path, 'c') # was 'cfu' create, fast, unlocked (TODO unlocked?!) but that only works if the gnu backend is used
		self.entries: Dict[bytes,PersistentEntry] = {}

	def __enter__(self):
		self._backing = self.db.__enter__()
		return self

	def __exit__(self, *args):
		self._backing.__exit__(*args)

	def __contains__(self, key: bytes):
		return key in self.entries or key in self._backing

	def __getitem__(self, key: bytes) -> PersistentEntry:
		if key not in self.entries:
			score = self._decode(self._backing[key])
			self.entries[key] = PersistentEntry(self, key, score)
		return self.entries[key]

	def __setitem__(self, key: bytes, value: Entry):
		self.entries[key] = PersistentEntry(self, key, value.score)

	def _write(self, key: bytes, value: float):
		self._backing[key] = self._encode(value)

	def _encode(self, value: float) -> bytes:
		return struct.pack('<f', value)

	def _decode(self, data: bytes) -> float:
		return struct.unpack('<f', data)[0]


T = TypeVar("T")


def none_throws(optional: Optional[T], message: str = "Unexpected `None`") -> T:
	"""Runtime cast of `Optional[T]` into `T`. Will raise an AssertionError if
	the argument was indeed `None`.
	"""
	if optional is None:
		raise ValueError(message)
	return optional


def exit_on_throw(fn):
	"""Wraps thread main function so that an exception thrown in the thread
	will terminate the entire process.
	"""
	@wraps(fn)
	def wrapper(*args, **kwargs):
		try:
			return fn(*args, **kwargs)
		except:
			print_exc(file=sys.stderr)
			os.kill(os.getpid(), signal.SIGKILL)
	return wrapper


def feed_child(queue, fin, fchild, cache):
	"""Thread that reads each line from stream `fin`, and will put its score
	`Entry` onto queue `queue`. If a line is a duplicate, it will use the entry
	from the previous occurrence. If a line is new, it will also feed it to
	child process `fchild` so a score can be calculated. Because the order of
	the queue and the order of feeding fchild are the same, the
	`threshold_scores` thread will know how to link them back together.
	"""
	derive_key = lambda val: xxh32(val).digest()

	try:
		for line in fin:
			key = derive_key(line)
			# if key not in cache, we've never seen the sentence
			if key not in cache:
				fchild.write(line)
				cache[key] = Entry()

			queue.put((line, cache[key]))
		fchild.close()
	except BrokenPipeError:
		pass
	finally:
		queue.put(None) # End indicator
		fin.close()


def threshold_scores(queue, fchild, fout, threshold, operator):
	"""Thread that reads the queue and, depending on the threshold, will write
	the line to output. It will also read any missing scores from the child
	`fchild`. Because this is the only thread reading & writing to Entry objects
	no locks are necessary.
	"""
	try:
		while True:
			item = queue.get()

			# Poison
			if item is None:
				break

			# If no score yet, get it from the child
			if item[1].score is None:
				item[1].score = float(fchild.readline())

			# If no threshold is specified, print everything and prefix it with the score
			if threshold is None:
				fout.write(str(item[1].score).encode() + b'\t' + item[0])
			
			# Otherwise only print the actual line if threshold is met
			elif operator(item[1].score, threshold):
				fout.write(item[0])
		
		# TODO: test somehow that child has stopped producing? Reading from `fchild`
		# should at this point return EOF since its stdin is already closed.
		fout.close()
	except BrokenPipeError:
		pass
	finally:
		fchild.close()


def open_cache(path: Optional[str]) -> Cache:
	"""Instantiates a cache type based on the path (or None) given."""
	if path:
		return PersistentCache(path)
	else:
		return Cache()


def main():
	parser = argparse.ArgumentParser(description=__doc__)
	parser.add_argument('threshold', type=float, help='Threshold (b) to compare score to.')
	parser.add_argument('scorer', type=str, nargs='+', help='Scorer program (a) and arguments.')
	parser.add_argument('--cache', '-c', type=str, help='Path to cache database.')
	
	ops = parser.add_mutually_exclusive_group()
	ops.set_defaults(operator=operator.ge) # default to --ge
	for name in ['lt', 'le', 'eq', 'ne', 'ge', 'gt']:
		ops.add_argument(f'--{name}', dest='operator', action='store_const', const=getattr(operator, name), help=getattr(operator, name).__doc__)

	args, scorer_args = parser.parse_known_args()

	# TODO: Make this Popen call only necessary if there was any need for it,
	# i.e. not all sentences could be scored by just the cache. I'm tempted to
	# add yet another wrapper program that only starts the process once input
	# is readable from stdin and then just re-attaches stdin to the child? Bit
	# like how inetd works. Or should this be a task for the downstream scorer
	# i.e. only load the model once input is received?
	child = Popen(args.scorer + scorer_args, stdin=PIPE, stdout=PIPE)

	queue = SimpleQueue() # type: SimpleQueue[tuple[bytes,Entry]]

	try:
		with open_cache(args.cache) as cache:
			# Reads stdin, writes it to queue, and possibly to child for scoring.
			feeder = Thread(target=exit_on_throw(feed_child), args=[queue, sys.stdin.buffer, child.stdin, cache])
			feeder.start()

			# Reads queue, writes to stdout, reading scores from child if necessary.
			consumer = Thread(target=exit_on_throw(threshold_scores), args=[queue, child.stdout, sys.stdout.buffer, args.threshold, args.operator])
			consumer.start()

			# Feeder will be done at this point
			feeder.join()

			# Consumer will be done once it read the last None from the queue.
			consumer.join()

			# Feeder will close child.stdin when all input is processed, which should
			# cause child to terminate.
	except:
		none_throws(child.stdin).close()
	finally:
		sys.stderr.close()
		retval = child.wait()
		sys.exit(retval)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
import sys
from typing import NamedTuple, Dict


class SentencePair(NamedTuple):
	src: str
	trg: str
	rank: float

best: Dict[str,SentencePair] = {}

for line in sys.stdin:
	src, trg, checksum, rank = line.rstrip('\r\n').split('\t')

	if checksum not in best or best[checksum].rank < float(rank):
		best[checksum] = SentencePair(src, trg, float(rank))

for pair in best.values():
	print(f"{pair.src}\t{pair.trg}", file=sys.stdout)



#!/usr/bin/env python3

import sys
import time
from typing import List, Tuple, Iterable, TypeVar, Optional, TextIO
import argparse
import numpy as np
from laserembeddings import Laser
from numpy.linalg import norm
from numpy.polynomial import Polynomial
from collections import deque
from io import TextIOBase


def _compute_similarity(laser: Laser, batch: List[Tuple[str, str]], src_lang: str, tgt_lang: str) -> List[float]:
    assert len(batch) > 0
    embeddings_src = laser.embed_sentences([line[0] for line in batch], lang=src_lang)
    embeddings_tgt = laser.embed_sentences([line[1] for line in batch], lang=tgt_lang)
    return [float(sim) for sim in _cosine_sim(embeddings_src, embeddings_tgt)]


def _cosine_sim(emb1: np.ndarray, emb2: np.ndarray) -> np.ndarray:
    return np.sum(emb1 * emb2, axis=-1) / (norm(emb1, axis=-1) * norm(emb2, axis=-1))


def interpolate(sample: Iterable[Tuple[int, float]], target:float) -> int:
    poly = Polynomial.fit([duration for size, duration in sample], [size for size, duration in sample], 1)
    return int(poly(target)), poly


class NullIO(TextIOBase):
    """TextIO that does nothing, as if writing to /dev/null."""
    def write(self, data:str) -> int:
        return len(data)


T = TypeVar('T')

def chunked(iterable: Iterable[T], *, chunk_size:Optional[int]=None, chunk_time:Optional[float]=None, verbose:Optional[TextIO]=NullIO()) -> Iterable[List[T]]:
    """Self-tuning batching iterator"""
    it = iter(iterable)

    # Initial set of measurements we then interpolate from
    limit_samples = iter([32, 64, 128, 256, 512, 1024])

    # Chunk size limit for the next chunk
    limit = chunk_size or next(limit_samples)

    # Measurements (limited to most recent 32)
    measurements = deque([], maxlen=32)

    while True:
        # Create a chunk
        chunk = [el for _, el in zip(range(limit), it)]

        # Did we reach the end because the last read was accidentally
        # exactly the remainder of the dataset? Or because there was no
        # input to begin with?
        if not chunk:
            return

        # Measure how long it takes before we are asked for the next chunk
        yield_time = time.monotonic()
        yield chunk
        if len(chunk) < limit:
            break
        yield_duration = time.monotonic() - yield_time
        print(f"Chunk size {limit} took {yield_duration}s", file=verbose)
        measurements.append((limit, yield_duration))

        # If we're running in dynamic mode, update the chunk size limit
        if chunk_size is None and chunk_time is not None:
            try:
                # Next limit for sampling?
                limit = next(limit_samples)
            except StopIteration:
                # No, we've run all the samples. Use previous measurements
                limit, poly = interpolate(measurements, chunk_time)
                print(f'Fitted {poly}', file=verbose)

            print(f"Setting chunk size to {limit}", file=verbose)


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description="Filter a parallel dataset using LASER.")
    parser.add_argument("--verbose", action="store_true", help="Print tuning info")
    parser.add_argument("--batch-size", type=int, help="LASER batch size")
    parser.add_argument("--batch-latency", type=float, default=10.0, help="Tune batch size to process a batch every N seconds (defaults to 10s, ignored if --batch-size is given)")
    parser.add_argument("--src-lang", type=str, required=True, help="Two-letter source language code (ISO 639-1)")
    parser.add_argument("--tgt-lang", type=str, required=True, help="Two-letter target language code (ISO 639-1)")
    
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--threshold", type=float, help="Minimum accepted LASER score.")
    group.add_argument("--scores", action="store_true", help="Print scores instead of lines")

    args = parser.parse_args()

    if not args.scores and args.threshold is None:
        print("Either use --threshold or --scores", file=sys.stderr)

    laser = Laser()

    for batch in chunked(sys.stdin, chunk_size=args.batch_size, chunk_time=args.batch_latency, verbose=sys.stderr if args.verbose else NullIO()):
        # TODO error checking of column count?
        scores = _compute_similarity(laser, [tuple(line.rstrip("\r\n").split("\t")[:2]) for line in batch], args.src_lang, args.tgt_lang)

        if args.scores:
            for score in scores:
                print(score)
        else:
            for line, score in zip(batch, scores):
                if score >= args.threshold:
                    sys.stdout.write(line)


if __name__ == "__main__":
    main()